import typer
from loguru import logger

from .. import settings
from ..job_queue import JobQueueManager  # Adjust import as needed
from ..utils.logging import setup_logging
from .utils import parse_dict_or_list_param

# Create a Typer app for job queue management commands
app = typer.Typer(help="Job queue management commands")

setup_logging(level=settings.LOG_LEVEL)


@app.command()
def start_worker(
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    background: bool = typer.Option(
        False, "--background", "-b", help="Run the worker in the background"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
    num_workers: int | None = typer.Option(
        None,
        "--num-workers",
        "-n",
        help="Number of worker processes to start (pool mode)",
    ),
):
    """
    Start a worker or worker pool to process jobs.

    This command starts a worker process (or a pool of worker processes) that will
    execute jobs from the queue. The worker will continue running until stopped
    or can be run in the background.

    Args:
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        background: Run the worker in the background
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)
        num_workers: Number of worker processes to start (pool mode)

    Examples:
        # Start a worker with default settings
        $ flowerpower job-queue start-worker

        # Start a worker for a specific backend type
        $ flowerpower job-queue start-worker --type rq

        # Start a worker pool with 4 processes
        $ flowerpower job-queue start-worker --num-workers 4

        # Run a worker in the background
        $ flowerpower job-queue start-worker --background

        # Set a specific logging level
        $ flowerpower job-queue start-worker --log-level debug
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if num_workers:
            num_workers = worker.cfg.backend.num_workers

        if num_workers and num_workers > 1:
            worker.start_worker_pool(num_workers=num_workers, background=background)
        else:
            worker.start_worker(background=background)


@app.command()
def start_scheduler(
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    background: bool = typer.Option(
        False, "--background", "-b", help="Run the scheduler in the background"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
    interval: int = typer.Option(
        60, "--interval", "-i", help="Interval for checking jobs in seconds (RQ only)"
    ),
):
    """
    Start the scheduler process for queued jobs.

    This command starts a scheduler that manages queued jobs and scheduled tasks.
    Note that this is only needed for RQ workers, as APScheduler workers have
    their own built-in scheduler.

    Args:
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        background: Run the scheduler in the background
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)
        interval: Interval for checking jobs in seconds (RQ only)

    Examples:
        # Start a scheduler with default settings
        $ flowerpower job-queue start-scheduler

        # Start a scheduler for a specific backend type
        $ flowerpower job-queue start-scheduler --type rq

        # Run a scheduler in the background
        $ flowerpower job-queue start-scheduler --background

        # Set a specific scheduler check interval (RQ only)
        $ flowerpower job-queue start-scheduler --interval 30
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if worker.cfg.backend.type != "rq":
            logger.info(
                f"No scheduler needed for {worker.cfg.backend.type} workers. Skipping."
            )
            return

        worker.start_scheduler(background=background, interval=interval)


# @app.command()
# def cancel_all_jobs(
#     type: str | None = None,
#     queue_name: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Cancel all jobs from the scheduler.

#     Note: This is different from deleting jobs as it only stops them from running but keeps their history.

#     Args:
#         type: Type of the job queue (rq, apscheduler)
#         queue_name: Name of the queue (RQ only)
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         if worker.cfg.backend.type != "rq":
#             logger.info(f"Job cancellation is not supported for {worker.cfg.backend.type} workers. Skipping.")
#             return

#         worker.cancel_all_jobs(queue_name=queue_name)

# @app.command()
# def cancel_all_schedules(
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Cancel all schedules from the scheduler.

#     Note: This is different from deleting schedules as it only stops them from running but keeps their configuration.

#     Args:
#         type: Type of the job queue (rq, apscheduler)
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         worker.cancel_all_schedules()


@app.command()
def cancel_job(
    job_id: str = typer.Argument(..., help="ID of the job to cancel"),
    all: bool = typer.Option(
        False, "--all", "-a", help="Cancel all jobs instead of a specific one"
    ),
    queue_name: str | None = typer.Option(
        None,
        help="Name of the queue (RQ only). If provided with --all, cancels all jobs in the queue",
    ),
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
):
    """
    Cancel a job or multiple jobs in the queue.

    This command stops a job from executing (if it hasn't started yet) or signals
    it to stop (if already running). Canceling is different from deleting as it
    maintains the job history but prevents execution.

    Args:
        job_id: ID of the job to cancel (ignored if --all is used)
        all: Cancel all jobs instead of a specific one
        queue_name: For RQ only, specifies the queue to cancel jobs from
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)

    Examples:
        # Cancel a specific job
        $ flowerpower job-queue cancel-job job-123456

        # Cancel all jobs in the default queue
        $ flowerpower job-queue cancel-job --all dummy-id

        # Cancel all jobs in a specific queue (RQ only)
        $ flowerpower job-queue cancel-job --all dummy-id --queue-name high-priority

        # Specify the backend type explicitly
        $ flowerpower job-queue cancel-job job-123456 --type rq
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if worker.cfg.backend.type != "rq":
            logger.info(
                f"Job cancellation is not supported for {worker.cfg.backend.type} workers. Skipping."
            )
            return
        if all:
            count = worker.cancel_all_jobs(
                queue_name=queue_name if worker.cfg.backend.type == "rq" else None
            )
            logger.info(
                f"Cancelled {count} jobs"
                + (f" in queue '{queue_name}'" if queue_name else "")
            )
        else:
            worker.cancel_job(job_id)
            logger.info(f"Job {job_id} cancelled")


@app.command()
def cancel_schedule(
    schedule_id: str,
    all: bool = False,
    type: str | None = None,
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
    log_level: str = "info",
):
    """
    Cancel a specific schedule.

    Note: This is different from deleting a schedule as it only stops it from running but keeps its configuration.

    Args:
        schedule_id: ID of the schedule to cancel
        all: If True, cancel all schedules
        type: Type of the job queue (rq, apscheduler)
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if all:
            worker.cancel_all_schedules()
        else:
            worker.cancel_schedule(schedule_id)


# @app.command()
# def delete_all_jobs(
#     type: str | None = None,
#     queue_name: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Delete all jobs from the scheduler. Note that this is different from cancelling jobs
#     as it also removes job history and results.

#     Args:
#         queue_name: Name of the queue (RQ only)
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         worker.delete_all_jobs(queue_name=queue_name if worker.cfg.backend.type == "rq" else None)

# @app.command()
# def delete_all_schedules(
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Delete all schedules from the scheduler.

#     Args:
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         worker.delete_all_schedules()


@app.command()
def delete_job(
    job_id: str,
    all: bool = False,
    queue_name: str | None = None,
    type: str | None = None,
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
    log_level: str = "info",
):
    """
    Delete a specific job.

    Args:
        job_id: ID of the job to delete
        all: If True, delete all jobs
        queue_name: Name of the queue (RQ only). If provided and all is True, delete all jobs in the queue
        type: Type of the job queue (rq, apscheduler)
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if all:
            worker.delete_all_jobs(
                queue_name=queue_name if worker.cfg.backend.type == "rq" else None
            )
        else:
            worker.delete_job(job_id)


@app.command()
def delete_schedule(
    schedule_id: str,
    all: bool = False,
    type: str | None = None,
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
    log_level: str = "info",
):
    """
    Delete a specific schedule.

    Args:
        schedule_id: ID of the schedule to delete
        all: If True, delete all schedules
        type: Type of the job queue (rq, apscheduler)
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if all:
            worker.delete_all_schedules()
        else:
            worker.delete_schedule(schedule_id)


# @app.command()
# def get_job(
#     job_id: str,
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Get information about a specific job.

#     Args:
#         job_id: ID of the job
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         # show_jobs should display the job info
#         worker.show_jobs(job_id=job_id)

# @app.command()
# def get_job_result(
#     job_id: str,
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
#     wait: bool = True,
# ):
#     """
#     Get the result of a specific job.

#     Args:
#         job_id: ID of the job
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#         wait: Wait for the result if job is still running (APScheduler only)
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         # worker's get_job_result method will handle the result display
#         worker.get_job_result(job_id, wait=wait if worker.cfg.backend.type == "apscheduler" else False)

# @app.command()
# def get_jobs(
#     type: str | None = None,
#     queue_name: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     List all jobs.

#     Args:
#         queue_name: Name of the queue (RQ only)
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         worker.show_jobs()

# @app.command()
# def get_schedule(
#     schedule_id: str,
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Get information about a specific schedule.

#     Args:
#         schedule_id: ID of the schedule
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         # show_schedule should display the schedule info
#         worker.show_schedules(schedule_id=schedule_id)

# @app.command()
# def get_schedules(
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     List all schedules.

#     Args:
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         worker.show_schedules()


@app.command()
def show_job_ids(
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
):
    """
    Show all job IDs in the job queue.

    This command displays all job IDs currently in the system, helping you identify
    jobs for other operations like getting results, canceling, or deleting jobs.

    Args:
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)

    Examples:
        # Show job IDs using default settings
        $ flowerpower job-queue show-job-ids

        # Show job IDs for a specific queue type
        $ flowerpower job-queue show-job-ids --type rq

        # Show job IDs with a custom scheduler configuration
        $ flowerpower job-queue show-job-ids --name my-scheduler

        # Show job IDs with debug logging
        $ flowerpower job-queue show-job-ids --log-level debug
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        # worker's job_ids property will print the IDs
        ids = worker.job_ids
        # Ensure we always print something meaningful
        if not ids:
            logger.info("No job IDs found")
        # If the worker's property doesn't already print the IDs, print them here
        elif not isinstance(ids, type(None)):  # Check if None was returned
            for job_id in ids:
                print(f"- {job_id}")


@app.command()
def show_schedule_ids(
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
):
    """
    Show all schedule IDs in the job queue.

    This command displays all schedule IDs currently in the system, helping you
    identify schedules for other operations like pausing, resuming, or deleting schedules.

    Args:
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)

    Examples:
        # Show schedule IDs using default settings
        $ flowerpower job-queue show-schedule-ids

        # Show schedule IDs for a specific queue type
        $ flowerpower job-queue show-schedule-ids --type apscheduler

        # Show schedule IDs with a custom scheduler configuration
        $ flowerpower job-queue show-schedule-ids --name my-scheduler

        # Show schedule IDs with debug logging
        $ flowerpower job-queue show-schedule-ids --log-level debug
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        # worker's schedule_ids property will print the IDs
        ids = worker.schedule_ids
        # Ensure we always print something meaningful
        if not ids:
            logger.info("No schedule IDs found")
        # If the worker's property doesn't already print the IDs, print them here
        elif not isinstance(ids, type(None)):  # Check if None was returned
            for schedule_id in ids:
                print(f"- {schedule_id}")


# @app.command()
# def pause_all_schedules(
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Pause all schedules.

#     Note: This functionality is only available for APScheduler workers.

#     Args:
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         if worker.cfg.backend.type != "apscheduler":
#             logger.info(f"Schedule pausing is not supported for {worker.cfg.backend.type} workers.")
#             return
#         worker.pause_all_schedules()


@app.command()
def pause_schedule(
    schedule_id: str = typer.Argument(..., help="ID of the schedule to pause"),
    all: bool = typer.Option(
        False, "--all", "-a", help="Pause all schedules instead of a specific one"
    ),
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
):
    """
    Pause a schedule or multiple schedules.

    This command temporarily stops a scheduled job from running while maintaining its
    configuration. Paused schedules can be resumed later. Note that this functionality
    is only available for APScheduler workers.

    Args:
        schedule_id: ID of the schedule to pause (ignored if --all is used)
        all: Pause all schedules instead of a specific one
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)

    Examples:
        # Pause a specific schedule
        $ flowerpower job-queue pause-schedule schedule-123456

        # Pause all schedules
        $ flowerpower job-queue pause-schedule --all dummy-id

        # Specify the backend type explicitly
        $ flowerpower job-queue pause-schedule schedule-123456 --type apscheduler
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if worker.cfg.backend.type != "apscheduler":
            logger.info(
                f"Schedule pausing is not supported for {worker.cfg.backend.type} workers."
            )
            return
        if all:
            count = worker.pause_all_schedules()
            logger.info(f"Paused {count} schedules")
        else:
            success = worker.pause_schedule(schedule_id)
            if success:
                logger.info(f"Schedule {schedule_id} paused successfully")
            else:
                logger.error(f"Failed to pause schedule {schedule_id}")


# @app.command()
# def resume_all_schedules(
#     type: str | None = None,
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     log_level: str = "info",
# ):
#     """
#     Resume all paused schedules.

#     Note: This functionality is only available for APScheduler workers.

#     Args:
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         log_level: Logging level
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with JobQueueManager(
#         type=type, name=name, base_dir=base_dir, storage_options=parsed_storage_options, log_level=log_level
#     ) as worker:
#         if worker.cfg.backend.type != "apscheduler":
#             logger.info(f"Schedule resuming is not supported for {worker.cfg.backend.type} workers.")
#             return
#         worker.resume_all_schedules()


@app.command()
def resume_schedule(
    schedule_id: str = typer.Argument(..., help="ID of the schedule to resume"),
    all: bool = typer.Option(
        False, "--all", "-a", help="Resume all schedules instead of a specific one"
    ),
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
):
    """
    Resume a paused schedule or multiple schedules.

    This command restarts previously paused schedules, allowing them to run again according
    to their original configuration. Note that this functionality is only available for
    APScheduler workers.

    Args:
        schedule_id: ID of the schedule to resume (ignored if --all is used)
        all: Resume all schedules instead of a specific one
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)

    Examples:
        # Resume a specific schedule
        $ flowerpower job-queue resume-schedule schedule-123456

        # Resume all schedules
        $ flowerpower job-queue resume-schedule --all dummy-id

        # Specify the backend type explicitly
        $ flowerpower job-queue resume-schedule schedule-123456 --type apscheduler

        # Set a specific logging level
        $ flowerpower job-queue resume-schedule schedule-123456 --log-level debug
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        if worker.cfg.backend.type != "apscheduler":
            logger.info(
                f"Schedule resuming is not supported for {worker.cfg.backend.type} workers."
            )
            return
        if all:
            count = worker.resume_all_schedules()
            logger.info(f"Resumed {count} schedules")
        else:
            success = worker.resume_schedule(schedule_id)
            if success:
                logger.info(f"Schedule {schedule_id} resumed successfully")
            else:
                logger.error(f"Failed to resume schedule {schedule_id}")


@app.command()
def show_jobs(
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    queue_name: str | None = typer.Option(
        None, help="Name of the queue to show jobs from (RQ only)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
    format: str = typer.Option("table", help="Output format (table, json, yaml)"),
):
    """
    Display detailed information about all jobs in the queue.

    This command shows comprehensive information about jobs including their status,
    creation time, execution time, and other details in a user-friendly format.

    Args:
        type: Type of job queue backend (rq, apscheduler)
        queue_name: Name of the queue to show jobs from (RQ only)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)
        format: Output format for the job information

    Examples:
        # Show all jobs using default settings
        $ flowerpower job-queue show-jobs

        # Show jobs for a specific queue type
        $ flowerpower job-queue show-jobs --type rq

        # Show jobs in a specific RQ queue
        $ flowerpower job-queue show-jobs --queue-name high-priority

        # Display jobs in JSON format
        $ flowerpower job-queue show-jobs --format json
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        worker.show_jobs(queue_name=queue_name, format=format)


@app.command()
def show_schedules(
    type: str | None = typer.Option(
        None, help="Type of job queue backend (rq, apscheduler)"
    ),
    name: str | None = typer.Option(
        None, help="Name of the scheduler configuration to use"
    ),
    base_dir: str | None = typer.Option(
        None, help="Base directory for the scheduler configuration"
    ),
    storage_options: str | None = typer.Option(
        None, help="Storage options as JSON or key=value pairs"
    ),
    log_level: str = typer.Option(
        "info", help="Logging level (debug, info, warning, error, critical)"
    ),
    format: str = typer.Option("table", help="Output format (table, json, yaml)"),
):
    """
    Display detailed information about all schedules.

    This command shows comprehensive information about scheduled jobs including their
    timing configuration, status, and other details in a user-friendly format.

    Args:
        type: Type of job queue backend (rq, apscheduler)
        name: Name of the scheduler configuration to use
        base_dir: Base directory for the scheduler configuration
        storage_options: Storage options as JSON or key=value pairs
        log_level: Logging level (debug, info, warning, error, critical)
        format: Output format for the schedule information

    Examples:
        # Show all schedules using default settings
        $ flowerpower job-queue show-schedules

        # Show schedules for a specific queue type
        $ flowerpower job-queue show-schedules --type apscheduler

        # Display schedules in JSON format
        $ flowerpower job-queue show-schedules --format json
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with JobQueueManager(
        type=type,
        name=name,
        base_dir=base_dir,
        storage_options=parsed_storage_options,
        log_level=log_level,
    ) as worker:
        worker.show_schedules(format=format)



import datetime as dt
import random
import socket
import time
from pathlib import Path
from types import TracebackType
from typing import Any, Callable

import mmh3
from loguru import logger
from munch import Munch
from paho.mqtt.client import (MQTT_ERR_SUCCESS, CallbackAPIVersion, Client,
                              MQTTMessageInfo)
from paho.mqtt.reasoncodes import ReasonCode

from ...cfg import ProjectConfig
from ...cfg.pipeline.run import ExecutorConfig, WithAdapterConfig
from ...cfg.project.adapter import AdapterConfig
from ...fs import AbstractFileSystem, BaseStorageOptions, get_filesystem
from ...pipeline.manager import PipelineManager
from ...utils.logging import setup_logging
from .cfg import MqttConfig

setup_logging()


class MqttManager:
    def __init__(
        self,
        username: str | None = None,
        password: str | None = None,
        host: str | None = "localhost",
        port: int | None = 1883,
        topic: str | None = None,
        first_reconnect_delay: int = 1,
        max_reconnect_count: int = 5,
        reconnect_rate: int = 2,
        max_reconnect_delay: int = 60,
        transport: str = "tcp",
        clean_session: bool = True,
        client_id: str | None = None,
        client_id_suffix: str | None = None,
        **kwargs,
    ):
        if "user" in kwargs:
            username = kwargs["user"]
        if "pw" in kwargs:
            password = kwargs["pw"]

        self.topic = topic

        self._username = username
        self._password = password
        self._host = host
        self._port = port
        self._first_reconnect_delay = first_reconnect_delay
        self._max_reconnect_count = max_reconnect_count
        self._reconnect_rate = reconnect_rate
        self._max_reconnect_delay = max_reconnect_delay
        self._transport = transport

        self._clean_session = clean_session
        self._client_id = client_id
        self._client_id_suffix = client_id_suffix

        self._client = None

    @classmethod
    def from_event_broker(cls, base_dir: str | None = None):
        """Create a MqttManager instance from the event broker configuration.

        Args:
            base_dir (str | None): Base directory for the project. If None, uses the current working directory.

        Returns:
            MqttManager: An instance of MqttManager configured with the event broker settings.
        """
        base_dir = base_dir or str(Path.cwd())

        jq_backend = ProjectConfig.load(base_dir=base_dir).job_queue.backend
        if jq_backend is None:
            raise ValueError(
                "No MQTT event broker configuration found. Recheck the provided `base_dir`.\n"
                "If you are not using a MQTT event broker, initialize the MQTT client using the MQTTManager class.\n"
                "or use the provide a configuration dict to the MQTTManager.from_dict() method."
            )
        if hasattr(jq_backend, "event_broker") is False:
            raise ValueError(
                "No MQTT event broker configuration found. Recheck the provided `base_dir`.\n"
                "If you are not using a MQTT event broker, initialize the MQTT client using the MQTTManager class.\n"
                "or use the provide a configuration dict to the MQTTManager.from_dict() method."
            )
        if jq_backend.event_broker.type != "mqtt":
            raise ValueError(
                "No MQTT event broker configuration found. Recheck the provided `base_dir`.\n"
                "If you are not using a MQTT event broker, initialize the MQTT client using the MQTTManager class.\n"
                "or use the provide a configuration dict to the MQTTManager.from_dict() method."
            )
        else:
            event_broker_cfg = jq_backend.event_broker
            return cls(
                **event_broker_cfg.dict(),
            )

    @classmethod
    def from_config(
        cls,
        cfg: MqttConfig | None = None,
        path: str | None = None,
        fs: AbstractFileSystem | None = None,
        storage_options: dict | BaseStorageOptions = {},
    ):
        """Create a MqttManager instance from the provided configuration.

        Args:
            cfg (MqttConfig | None): Configuration for the MQTT client. If None, loads from the provided path.
            path (str | None): Path to the configuration file. If None, uses the default configuration.
            fs (AbstractFileSystem | None): File system to use for loading the configuration.
            storage_options (dict | BaseStorageOptions): Storage options for the file system.

        Returns:
            MqttManager: An instance of MqttManager configured with the provided settings.
        """
        if cfg is None:
            if path is None:
                raise ValueError(
                    "No configuration provided. Please provide `config` or `path` to the configuration file."
                )

        if cfg is None:
            import os

            if fs is None:
                fs = get_filesystem(
                    path=os.path.dirname(path), storage_options=storage_options
                )

            cfg = MqttConfig.from_yaml(path=os.path.basename(path), fs=fs)

        return cls(
            **cfg.dict(),
        )

    @classmethod
    def from_dict(cls, cfg: dict):
        """Create a MqttManager instance from the provided dictionary configuration.

        Args:
            cfg (dict): Dictionary containing the configuration for the MQTT client.

        Returns:
            MqttManager: An instance of MqttManager configured with the provided settings.
        """
        return cls(
            **cfg,
        )

    def __enter__(self) -> "MqttManager":
        self.connect()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        # Add any cleanup code here if needed
        self.disconnect()

    @staticmethod
    def _on_connect(client, userdata, flags, rc, properties):
        if rc == 0:
            logger.info(f"Connected to MQTT Broker {userdata.host}!")
            logger.info(
                f"Connected as {userdata.client_id} with clean session {userdata.clean_session}"
            )
        else:
            logger.error(f"Failed to connect, return code {rc}")

    @staticmethod
    def _on_disconnect(client, userdata, disconnect_flags, rc, properties=None):
        reconnect_count, reconnect_delay = 0, userdata.first_reconnect_delay

        if userdata.max_reconnect_count == 0:
            logger.info("Disconnected successfully!")
            return

        while reconnect_count < userdata.max_reconnect_count:
            logger.info(f"Reconnecting in {reconnect_delay} seconds...")
            time.sleep(reconnect_delay)

            try:
                client.reconnect()
                logger.info("Reconnected successfully!")
                return
            except Exception as err:
                logger.error(f"{err}. Reconnect failed. Retrying...")

            reconnect_delay *= userdata.reconnect_rate
            reconnect_delay = min(reconnect_delay, userdata.max_reconnect_delay)
            reconnect_count += 1
        logger.info(f"Reconnect failed after {reconnect_count} attempts. Exiting...")

    # @staticmethod
    # def _on_publish(client, userdata, mid, rc, properties):
    #     logger.info(f"Published message id: {mid} with return code: {rc}")
    #     if rc == 0:
    #         logger.debug("Message published successfully!")
    #     else:
    #         logger.debug(f"Failed to publish message with return code: {rc}")

    @staticmethod
    def _on_publish(client, userdata, mid, reason_code_obj, properties):
        """Callback function for when a message is published."""
        if isinstance(reason_code_obj, ReasonCode):
            if not reason_code_obj.is_failure:
                logger.info(
                    f"Broker acknowledged message_id: {mid} (ReasonCode: {reason_code_obj})"
                )
            else:
                if reason_code_obj.value == 16:  # MQTTReasonCode.NoMatchingSubscribers
                    logger.warning(
                        f"Message_id: {mid} published, but broker reported no matching subscribers (ReasonCode: {reason_code_obj})."
                    )
                else:
                    logger.error(
                        f"Broker acknowledgment error for message_id: {mid}. ReasonCode: {reason_code_obj}"
                    )
        elif isinstance(reason_code_obj, int):  # Fallback for simpler acks (legacy RCs)
            if reason_code_obj == 0:  # MQTT_ERR_SUCCESS
                logger.info(
                    f"Broker acknowledged message_id: {mid} (Legacy RC: {reason_code_obj})"
                )
            else:
                logger.error(
                    f"Broker acknowledgment error for message_id: {mid}. Legacy RC: {reason_code_obj} ({client.error_string(reason_code_obj)})"
                )
        else:
            logger.warning(
                f"Message_id: {mid} published. Received unusual RC type in on_publish: {reason_code_obj} (Type: {type(reason_code_obj)})"
            )

    @staticmethod
    def _on_subscribe(client, userdata, mid, qos, properties):
        if isinstance(qos, list):
            qos_msg = str(qos[0])
        else:
            qos_msg = f"and granted QoS {qos[0]}"
        logger.info(f"Subscribed {qos_msg}")

    def connect(self) -> Client:
        """Connect to the MQTT broker.
        Returns:
            Client: The connected MQTT client.
        """
        if self._client_id is None and self._clean_session:
            # Random Client ID when clean session is True
            self._client_id = f"flowerpower-client-{random.randint(0, 10000)}"
        elif self._client_id is None and not self._clean_session:
            # Deterministic Client ID when clean session is False
            self._client_id = f"flowerpower-client-{
                mmh3.hash_bytes(
                    str(self._host)
                    + str(self._port)
                    + str(self.topic)
                    + str(socket.gethostname())
                ).hex()
            }"

        if self._client_id_suffix:
            self._client_id = f"{self._client_id}-{self._client_id_suffix}"

        logger.debug(
            f"Client ID: {self._client_id} - Clean session: {self._clean_session}"
        )
        client = Client(
            CallbackAPIVersion.VERSION2,
            client_id=self._client_id,
            transport=self._transport,
            clean_session=self._clean_session,
            userdata=Munch(
                user=self._username,
                pw=self._password,
                host=self._host,
                port=self._port,
                topic=self.topic,
                first_reconnect_delay=self._first_reconnect_delay,
                max_reconnect_count=self._max_reconnect_count,
                reconnect_rate=self._reconnect_rate,
                max_reconnect_delay=self._max_reconnect_delay,
                transport=self._transport,
                client_id=self._client_id,
                clean_session=self._clean_session,
            ),
        )
        if self._password != "" and self._username != "":
            client.username_pw_set(self._username, self._password)

        client.on_connect = self._on_connect  # self._on_connect
        client.on_disconnect = self._on_disconnect  # self._on_disconnect
        client.on_publish = self._on_publish
        client.on_subscribe = self._on_subscribe

        client.connect(self._host, self._port)
        self._client = client
        # topic = topic or topic
        if self.topic:
            self.subscribe()

    def disconnect(self):
        """Disconnect from the MQTT broker."""
        if self._client is None:
            logger.warning("Client is not connected. Cannot disconnect.")
            return
        self._max_reconnect_count = 0
        self._client._userdata.max_reconnect_count = 0
        self._client.disconnect()

    def reconnect(self):
        """Reconnect to the MQTT broker."""
        if self._client is None:
            logger.warning("Client is not connected. Connecting instead.")
            self.connect()
        else:
            self._client.reconnect()

    def publish(
        self, topic: str, payload: Any, qos: int = 0, retain: bool = False
    ) -> "MQTTMessageInfo | None":
        """
        Publish a message to the MQTT broker.

        Args:
            topic (str): The topic to publish the message to.
            payload (Any): The message payload.
            qos (int, optional): The Quality of Service level. Defaults to 0.
            retain (bool, optional): Whether to retain the message. Defaults to False.

        Returns:
            MQTTMessageInfo | None: Information about the published message, or None if an error occurred.
        """
        if self._client is None or not self._client.is_connected():
            logger.warning(
                "Client is not connected. Attempting to connect before publishing."
            )
            try:
                self.connect()
            except Exception as e:
                logger.error(f"Failed to connect to MQTT broker before publishing: {e}")
                return None

        try:
            msg_info = self._client.publish(
                topic=topic, payload=payload, qos=qos, retain=retain
            )
            if msg_info.rc == MQTT_ERR_SUCCESS:  # 0:
                logger.debug(
                    f"Message published successfully. MID: {msg_info.mid}, Topic: {topic}, QoS: {qos}, Retain: {retain}, RC: {msg_info.rc}"
                )
            else:
                logger.error(
                    f"Failed to publish message. Topic: {topic}, QoS: {qos}, Retain: {retain}, RC: {msg_info.rc}, Error: {self._client.error_string(msg_info.rc)}"
                )
            return msg_info
        except Exception as e:
            logger.error(
                f"An error occurred while publishing message to topic {topic}: {e}"
            )
            return None

    def subscribe(self, topic: str | None = None, qos: int = 2):
        """
        Subscribe to a topic on the MQTT broker.

        Args:
            topic (str | None): The topic to subscribe to. If None, uses the instance's topic.
            qos (int): The Quality of Service level.
        """
        if self._client is None or not self._client.is_connected():
            self.connect()
        if topic is not None:
            self.topic = topic
        self._client.subscribe(self.topic, qos=qos)

    def unsubscribe(self, topic: str | None = None):
        """
        Unsubscribe from a topic on the MQTT broker.

        Args:
            topic (str | None): The topic to unsubscribe from. If None, uses the instance's topic.
        """
        if self._client is None or not self._client.is_connected():
            self.connect()
        if topic is not None:
            self.topic = topic
        self._client.unsubscribe(self.topic)

    def register_on_message(self, on_message: Callable):
        """
        Register a callback function to be called when a message is received.

        Args:
            on_message (Callable): The callback function to register.
        """
        if self._client is None or not self._client.is_connected():
            self.connect()
        self._client.on_message = on_message

    def run_in_background(
        self,
        on_message: Callable,
        topic: str | None = None,
        qos: int = 2,
    ) -> None:
        """
        Run the MQTT client in the background.

        Args:
            on_message: Callback function to run when a message is received
            topic: MQTT topic to listen to

        Returns:
            None


        """
        if self._client is None or not self._client.is_connected():
            self.connect()

        if topic:
            self.subscribe(topic, qos=qos)

        self._client.on_message = on_message
        self._client.loop_start()

    def run_until_break(
        self,
        on_message: Callable,
        topic: str | None = None,
        qos: int = 2,
    ):
        """
        Run the MQTT client until a break signal is received.

        Args:
            on_message: Callback function to run when a message is received
            topic: MQTT topic to listen to

        Returns:
            None
        """
        if self._client is None or not self._client.is_connected():
            self.connect()

        if topic:
            self.subscribe(topic, qos=qos)

        self._client.on_message = on_message
        self._client.loop_forever()

    def start_listener(
        self,
        on_message: Callable,
        topic: str | None = None,
        background: bool = False,
        qos: int = 2,
    ) -> None:
        """
        Start the MQTT listener.

        Args:
            on_message: Callback function to run when a message is received
            topic: MQTT topic to listen to
            background: Run the listener in the background

        Returns:
            None
        """
        if background:
            self.run_in_background(on_message, topic, qos)
        else:
            self.run_until_break(on_message, topic, qos)

    def stop_listener(
        self,
    ) -> None:
        """
        Stop the MQTT listener.

        Returns:
            None
        """
        self._client.loop_stop()
        logger.info("Client stopped.")

    def run_pipeline_on_message(
        self,
        name: str,
        topic: str | None = None,
        inputs: dict | None = None,
        final_vars: list | None = None,
        config: dict | None = None,
        cache: bool | dict = False,
        executor_cfg: str | dict | ExecutorConfig | None = None,
        with_adapter_cfg: dict | WithAdapterConfig | None = None,
        pipeline_adapter_cfg: dict | AdapterConfig | None = None,
        project_adapter_cfg: dict | AdapterConfig | None = None,
        adapter: dict[str, Any] | None = None,
        reload: bool = False,
        log_level: str | None = None,
        result_ttl: float | dt.timedelta = 0,
        run_in: int | str | dt.timedelta | None = None,
        max_retries: int | None = None,
        retry_delay: float | None = None,
        jitter_factor: float | None = None,
        retry_exceptions: tuple | list | None = None,
        as_job: bool = False,
        base_dir: str | None = None,
        storage_options: dict = {},
        fs: AbstractFileSystem | None = None,
        background: bool = False,
        qos: int = 2,
        config_hook: Callable[[bytes, int], dict] | None = None,
        **kwargs,
    ):
        """
        Start a pipeline listener that listens to a topic and processes the message using a pipeline.

        Args:
            name (str): Name of the pipeline
            topic (str | None): MQTT topic to listen to
            inputs (dict | None): Inputs for the pipeline
            final_vars (list | None): Final variables for the pipeline
            config (dict | None): Configuration for the pipeline driver
            cache (bool | dict): Cache for the pipeline
            executor_cfg (str | dict | ExecutorConfig | None): Executor configuration
            with_adapter_cfg (dict | WithAdapterConfig | None): With adapter configuration
            pipeline_adapter_cfg (dict | AdapterConfig | None): Pipeline adapter configuration
            project_adapter_cfg (dict | AdapterConfig | None): Project adapter configuration
            adapter (dict[str, Any] | None): Adapter configuration
            reload (bool): Reload the pipeline
            log_level (str | None): Log level for the pipeline
            result_ttl (float | dt.timedelta): Result expiration time for the pipeline
            run_in (int | str | dt.timedelta | None): Run in time for the pipeline
            max_retries (int | None): Maximum number of retries for the pipeline
            retry_delay (float | None): Delay between retries for the pipeline
            jitter_factor (float | None): Jitter factor for the pipeline
            retry_exceptions (tuple | list | None): Exceptions to retry for the pipeline
            as_job (bool): Run the pipeline as a job
            base_dir (str | None): Base directory for the pipeline
            storage_options (dict): Storage options for the pipeline
            fs (AbstractFileSystem | None): File system for the pipeline
            background (bool): Run the listener in the background
            qos (int): Quality of Service for the MQTT client
            config_hook (Callable[[bytes, int], dict] | None): Hook function to modify the configuration of the pipeline
            **kwargs: Additional keyword arguments

        Returns:
            None

        Raises:
            ValueError: If the config_hook is not callable

        Example:
            ```python
            from flowerpower.plugins.mqtt import MqttManager
            mqtt = MqttManager()
            mqtt.run_pipeline_on_message(
                name="my_pipeline",
                topic="my_topic",
                inputs={"key": "value"},
                config={"param": "value"},
                as_job=True,
            )
            ```
        """

        if inputs is None:
            inputs = {}

        if config is None:
            config = {}

        if config_hook is not None and not callable(config_hook):
            raise ValueError("config_hook must be a callable function")

        def on_message(client, userdata, msg):
            # logger.info(f"Received message on topic {topic}")
            logger.info(
                f"Received message on subscribed topic {topic} (exact topic {msg.topic})"
            )

            inputs["payload"] = msg.payload
            inputs["topic"] = msg.topic

            if config_hook is not None:
                # config_ = config_hook(inputs["payload"], inputs["topic"])
                try:
                    config_ = config_hook(inputs["payload"], inputs["topic"])
                    logger.debug(f"Config from hook: {config_}")
                except Exception as e:
                    # _ = e
                    logger.warning("Config hook failed. Aborting Message processing")
                    logger.exception(e)
                    return

                if any([k in config_ for k in config.keys()]):
                    logger.warning("Config from hook overwrites config from pipeline")

                config.update(config_)
                logger.debug(f"Config after update: {config}")

            with PipelineManager(
                storage_options=storage_options, fs=fs, base_dir=base_dir
            ) as pipeline:
                try:
                    if as_job:
                        pipeline.add_job(
                            name=name,
                            inputs=inputs,
                            final_vars=final_vars,
                            config=config,
                            cache=cache,
                            executor_cfg=executor_cfg,
                            with_adapter_cfg=with_adapter_cfg,
                            pipeline_adapter_cfg=pipeline_adapter_cfg,
                            project_adapter_cfg=project_adapter_cfg,
                            adapter=adapter,
                            run_in=run_in,
                            reload=reload,
                            log_level=log_level,
                            result_ttl=result_ttl,
                            max_retries=max_retries,
                            retry_delay=retry_delay,
                            jitter_factor=jitter_factor,
                            retry_exceptions=retry_exceptions,
                            **kwargs,
                        )
                    else:
                        pipeline.run(
                            name=name,
                            inputs=inputs,
                            final_vars=final_vars,
                            config=config,
                            cache=cache,
                            executor_cfg=executor_cfg,
                            with_adapter_cfg=with_adapter_cfg,
                            pipeline_adapter_cfg=pipeline_adapter_cfg,
                            project_adapter_cfg=project_adapter_cfg,
                            adapter=adapter,
                            reload=reload,
                            log_level=log_level,
                            max_retries=max_retries,
                            retry_delay=retry_delay,
                            jitter_factor=jitter_factor,
                            retry_exceptions=retry_exceptions,
                        )
                    logger.success("Message processed successfully")
                    return
                except Exception as e:
                    _ = e
                    logger.exception(e)

                logger.warning("Message processing failed")

        self.start_listener(
            on_message=on_message, topic=topic, background=background, qos=qos
        )


def start_listener(
    on_message: Callable,
    topic: str | None = None,
    background: bool = False,
    mqtt_cfg: dict | MqttConfig = {},
    base_dir: str | None = None,
    username: str | None = None,
    password: str | None = None,
    host: str | None = None,
    port: int | None = None,
    clean_session: bool = True,
    qos: int = 2,
    client_id: str | None = None,
    client_id_suffix: str | None = None,
    config_hook: Callable[[bytes, int], dict] | None = None,
    **kwargs,
) -> None:
    """
    Start the MQTT listener.

    The connection to the MQTT broker is established using the provided configuration of a
    MQTT event broker defined in the project configuration file `conf/project.toml`.
    If no configuration is found, you have to provide either the argument `mqtt_cfg`, dict with the
    connection parameters or the arguments `username`, `password`, `host`, and `port`.

    Args:
        on_message (Callable): Callback function to run when a message is received
        topic (str | None): MQTT topic to listen to
        background (bool): Run the listener in the background
        mqtt_cfg (dict | MqttConfig): MQTT client configuration. Use either this or arguments
            username, password, host, and port.
        base_dir (str | None): Base directory for the module
        username (str | None): Username for the MQTT client
        password (str | None): Password for the MQTT client
        host (str | None): Host for the MQTT client
        port (int | None): Port for the MQTT client
        clean_session (bool): Clean session flag for the MQTT client
        qos (int): Quality of Service for the MQTT client
        client_id (str | None): Client ID for the MQTT client
        client_id_suffix (str | None): Client ID suffix for the MQTT client
        config_hook (Callable[[bytes, int], dict] | None): Hook function to modify the configuration of the pipeline
        **kwargs: Additional keyword arguments

    Returns:
        None

    Raises:
        ValueError: If the config_hook is not callable
        ValueError: If no client configuration is found

    Example:
        ```python
        from flowerpower.plugins.mqtt import start_listener

        start_listener(
            on_message=my_on_message_function,
            topic="my_topic",
            background=True,
            mqtt_cfg={"host": "localhost", "port": 1883},
        )
        ```
    """
    try:
        client = MqttManager.from_event_broker(base_dir)
    except ValueError:
        if mqtt_cfg:
            if isinstance(mqtt_cfg, MqttConfig):
                client = MqttManager.from_config(mqtt_cfg)
            elif isinstance(mqtt_cfg, dict):
                client = MqttManager.from_dict(mqtt_cfg)
        elif host and port:
            client = MqttManager(
                username=username,
                password=password,
                host=host,
                port=port,
                clean_session=clean_session,
                client_id=client_id,
                client_id_suffix=client_id_suffix,
                config_hook=config_hook,
                **kwargs,
            )
        else:
            raise ValueError(
                "No client configuration found. Please provide a client configuration "
                "or a FlowerPower project base directory, in which a event broker is "
                "configured in the `config/project.yml` file."
            )

    client.start_listener(
        on_message=on_message, topic=topic, background=background, qos=qos
    )


def run_pipeline_on_message(
    name: str,
    topic: str | None = None,
    inputs: dict | None = None,
    final_vars: list | None = None,
    config: dict | None = None,
    cache: bool | dict = False,
    executor_cfg: str | dict | ExecutorConfig | None = None,
    with_adapter_cfg: dict | WithAdapterConfig | None = None,
    pipeline_adapter_cfg: dict | AdapterConfig | None = None,
    project_adapter_cfg: dict | AdapterConfig | None = None,
    adapter: dict[str, Any] | None = None,
    reload: bool = False,
    log_level: str | None = None,
    result_ttl: float | dt.timedelta = 0,
    run_in: int | str | dt.timedelta | None = None,
    max_retries: int | None = None,
    retry_delay: float | None = None,
    jitter_factor: float | None = None,
    retry_exceptions: tuple | list | None = None,
    as_job: bool = False,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
    background: bool = False,
    mqtt_cfg: dict | MqttConfig = {},
    host: str | None = None,
    port: int | None = None,
    username: str | None = None,
    password: str | None = None,
    clean_session: bool = True,
    qos: int = 2,
    client_id: str | None = None,
    client_id_suffix: str | None = None,
    config_hook: Callable[[bytes, int], dict] | None = None,
    **kwargs,
):
    """
    Start a pipeline listener that listens to a topic and processes the message using a pipeline.

    Args:
        name (str): Name of the pipeline
        topic (str | None): MQTT topic to listen to
        inputs (dict | None): Inputs for the pipeline
        final_vars (list | None): Final variables for the pipeline
        config (dict | None): Configuration for the pipeline driver
        cache (bool | dict): Cache for the pipeline
        executor_cfg (str | dict | ExecutorConfig | None): Executor configuration
        with_adapter_cfg (dict | WithAdapterConfig | None): With adapter configuration
        pipeline_adapter_cfg (dict | AdapterConfig | None): Pipeline adapter configuration
        project_adapter_cfg (dict | AdapterConfig | None): Project adapter configuration
        adapter (dict[str, Any] | None): Adapter configuration
        reload (bool): Reload the pipeline
        log_level (str | None): Log level for the pipeline
        result_ttl (float | dt.timedelta): Result expiration time for the pipeline
        run_in (int | str | dt.timedelta | None): Run in time for the pipeline
        max_retries (int | None): Maximum number of retries for the pipeline
        retry_delay (float | None): Delay between retries for the pipeline
        jitter_factor (float | None): Jitter factor for the pipeline
        retry_exceptions (tuple | list | None): Exceptions to retry for the pipeline
        as_job (bool): Run the pipeline as a job
        base_dir (str | None): Base directory for the pipeline
        storage_options (dict): Storage options for the pipeline
        fs (AbstractFileSystem | None): File system for the pipeline
        background (bool): Run the listener in the background
        mqtt_cfg (dict | MqttConfig): MQTT client configuration. Use either this or arguments
            username, password, host, and port.
        host (str | None): Host for the MQTT client
        port (int | None): Port for the MQTT client
        username (str | None): Username for the MQTT client
        password (str | None): Password for the MQTT client
        clean_session (bool): Clean session flag for the MQTT client
        qos (int): Quality of Service for the MQTT client
        client_id (str | None): Client ID for the MQTT client
        client_id_suffix (str | None): Client ID suffix for the MQTT client
        config_hook (Callable[[bytes, int], dict] | None): Hook function to modify the configuration of the pipeline
        **kwargs: Additional keyword arguments

    Returns:
        None

    Raises:
        ValueError: If the config_hook is not callable
        ValueError: If no client configuration is found

    Example:
        ```python
        from flowerpower.plugins.mqtt import run_pipeline_on_message

        run_pipeline_on_message(
            name="my_pipeline",
            topic="my_topic",
            inputs={"key": "value"},
            config={"param": "value"},
            as_job=True,
        )
        ```
    """
    try:
        client = MqttManager.from_event_broker(base_dir)
    except ValueError:
        if mqtt_cfg:
            if isinstance(mqtt_cfg, MqttConfig):
                client = MqttManager.from_config(mqtt_cfg)
            elif isinstance(mqtt_cfg, dict):
                client = MqttManager.from_dict(mqtt_cfg)
        elif host and port:
            client = MqttManager(
                username=username,
                password=password,
                host=host,
                port=port,
                clean_session=clean_session,
                client_id=client_id,
                client_id_suffix=client_id_suffix,
                config_hook=config_hook,
                **kwargs,
            )
        else:
            raise ValueError(
                "No client configuration found. Please provide a client configuration "
                "or a FlowerPower project base directory, in which a event broker is "
                "configured in the `config/project.yml` file."
            )

    if client._client_id is None and client_id is not None:
        client._client_id = client_id

    if client._client_id_suffix is None and client_id_suffix is not None:
        client._client_id_suffix = client_id_suffix

    """
    cli_clean_session | config_clean_session | result
    TRUE		        TRUE		           TRUE
    FALSE		        FALSE                  FALSE
    FALSE               TRUE                   FALSE
    TRUE                FALSE                  FALSE

    Clean session should only use default value if neither cli nor config source says otherwise
    """
    client._clean_session = client._clean_session and clean_session

    if client.topic is None and topic is not None:
        client.topic = topic

    client.run_pipeline_on_message(
        name=name,
        topic=topic,
        inputs=inputs,
        final_vars=final_vars,
        config=config,
        cache=cache,
        executor_cfg=executor_cfg,
        with_adapter_cfg=with_adapter_cfg,
        pipeline_adapter_cfg=pipeline_adapter_cfg,
        project_adapter_cfg=project_adapter_cfg,
        adapter=adapter,
        reload=reload,
        log_level=log_level,
        result_ttl=result_ttl,
        run_in=run_in,
        max_retries=max_retries,
        retry_delay=retry_delay,
        jitter_factor=jitter_factor,
        retry_exceptions=retry_exceptions,
        as_job=as_job,
        base_dir=base_dir,
        storage_options=storage_options,
        fs=fs,
        background=background,
        qos=qos,
        client_id=client_id,
        client_id_suffix=client_id_suffix,
        config_hook=config_hook,
        **kwargs,
    )



# import datetime as dt


import datetime

import pyarrow as pa
import pyarrow.dataset as pds
from deltalake import DeltaTable, table
from deltalake.exceptions import TableNotFoundError
# from ..utils import get_dataframe_metadata, get_delta_metadata
from loguru import logger
from sherlock import RedisLock

from ..base import BaseDatasetReader
from ..metadata import (get_dataframe_metadata, get_delta_metadata,
                        get_pyarrow_dataset_metadata)

# from hamilton.function_modifiers import dataloader


class DeltaTableReader(BaseDatasetReader):
    """Delta table loader.

    This class is responsible for loading Delta tables into several dataframe formats,
    duckdb and datafusion.

    """

    delta_table: DeltaTable | None = None
    with_lock: bool = False
    redis: str | None = None
    format: str = "delta"

    def model_post_init(self, __context):
        super().model_post_init(__context)

        self._init_dt()
        if self.with_lock and self.redis is None:
            raise ValueError("Redis connection is required when using locks.")

    def _init_dt(self):
        try:
            self.delta_table = DeltaTable(
                self._raw_path,
                storage_options=self.storage_options.to_object_store_kwargs(),
            )
        except TableNotFoundError:
            logger.warning(f"Table {self._raw_path} not found.")
            self.delta_table = None

    @property
    def dt(self) -> DeltaTable:
        return self.delta_table

    def _load(self, reload: bool = False):
        self.to_pyarrow_table(reload=reload)

    def to_pyarrow_dataset(
        self, metadata: bool = False, reload: bool = False
    ) -> pds.Dataset | tuple[pds.Dataset, dict[str, any]]:
        """Converts the DeltaTable to a PyArrow Dataset.

        Args:
            metadata (bool, optional): Whether to include metadata. Defaults to False.
            reload (bool, optional): Whether to reload the dataset. Defaults to False.

        Returns:
            pds.Dataset | tuple[pds.Dataset, dict[str, any]]: PyArrow Dataset or tuple of PyArrow Dataset and metadata.
        """
        if self.delta_table is None:
            self._init_dt()
            if self.delta_table is None:
                return None

        if reload or not hasattr(self, "_dataset"):
            self._dataset = self.delta_table.to_pyarrow_dataset()
        if metadata:
            metadata = get_pyarrow_dataset_metadata(
                self._dataset, self._raw_path, "parquet"
            )
            return self._dataset, metadata
        return self._dataset

    def to_pyarrow_table(
        self, metadata: bool = False, reload: bool = False
    ) -> pa.Table | tuple[pa.Table, dict[str, any]]:
        """Converts the DeltaTable to a PyArrow Table.

        Args:
            metadata (bool, optional): Whether to include metadata. Defaults to False.
            reload (bool, optional): Whether to reload the table. Defaults to False.

        Returns:
            pa.Table | tuple[pa.Table, dict[str, any]]: PyArrow Table or tuple of PyArrow Table and metadata.
        """
        if self.delta_table is None:
            self._init_dt()
            if self.delta_table is None:
                return None

        if reload or not hasattr(self, "_data"):
            self._data = self.delta_table.to_pyarrow_table()
        if metadata:
            metadata = get_dataframe_metadata(table, self._raw_path, "parquet")
            return self._data, metadata
        return self._data

    def compact(
        self,
        partition_filters: list[tuple[str, str, any]] | None = None,
        target_size: int = None,
        max_concurrent_tasks: int = None,
        min_commit_interval: int | datetime.timedelta | None = None,
        writer_properties: table.WriterProperties = None,
        custom_metadata: dict[str, str] | None = None,
        post_commithook_properties: table.PostCommitHookProperties | None = None,
        commit_properties: table.CommitProperties | None = None,
    ) -> dict[str, any]:
        def _compact():
            self.delta_table.compact(
                partition_filters=partition_filters,
                target_size=target_size,
                max_concurrent_tasks=max_concurrent_tasks,
                min_commit_interval=min_commit_interval,
                writer_properties=writer_properties,
                custom_metadata=custom_metadata,
                post_commithook_properties=post_commithook_properties,
                commit_properties=commit_properties,
            )

        if self.with_lock:
            with RedisLock(
                lock_name=self._raw_path,
                namespace="flowerpower",
                client=self.redis,
                expire=10,
                timeout=5,
                retry_interval=0.1,
            ):
                _compact()
        else:
            _compact()

    def z_order(
        self,
        columns: list[str],
        partition_filters: list[tuple[str, str, any]] | None = None,
        target_size: int = None,
        max_concurrent_tasks: int = None,
        min_commit_interval: int | datetime.timedelta | None = None,
        writer_properties: table.WriterProperties = None,
        custom_metadata: dict[str, str] | None = None,
        post_commithook_properties: table.PostCommitHookProperties | None = None,
        commit_properties: table.CommitProperties | None = None,
    ) -> dict[str, any]:
        def _z_order():
            self.delta_table.z_order(
                columns=columns,
                partition_filters=partition_filters,
                target_size=target_size,
                max_concurrent_tasks=max_concurrent_tasks,
                min_commit_interval=min_commit_interval,
                writer_properties=writer_properties,
                custom_metadata=custom_metadata,
                post_commithook_properties=post_commithook_properties,
                commit_properties=commit_properties,
            )

        if self.with_lock:
            with RedisLock(
                lock_name=self._raw_path,
                namespace="flowerpower",
                client=self.redis,
                expire=10,
                timeout=5,
                retry_interval=0.1,
            ):
                _z_order()
        else:
            _z_order()

    @property
    def metadata(self) -> dict:
        if not hasattr(self, "_metadata"):
            self._metadata = get_delta_metadata(self.delta_table, self._raw_path)
        return self._metadata



from ..base import BaseDatabaseWriter


class SQLiteWriter(BaseDatabaseWriter):
    """SQLite writer.

    This class is responsible for writing dataframes to SQLite database.

    Examples:
        ```python
        writer = SQLiteWriter(table_name="table", path="data.db")
        writer.write(df)

        # or
        writer = SQLiteWriter(table_name="table",
                                connection_string="sqkite:///data.db")
        writer.write(df)
        ```
    """

    type_: str = "sqlite"

    def model_post_init(self, __context):
        super().model_post_init(__context)
        # self.type_ = "sqlite"



import importlib

has_orjson = importlib.util.find_spec("orjson") is not None
has_polars = importlib.util.find_spec("polars") is not None

if has_orjson and has_polars:
    from .ext import AbstractFileSystem
else:
    from fsspec import AbstractFileSystem

from .base import get_filesystem  # noqa: E402
from .storage_options import AwsStorageOptions  # noqa: E402
from .storage_options import AzureStorageOptions  # noqa: E402
from .storage_options import (BaseStorageOptions, GcsStorageOptions,
                              GitHubStorageOptions, GitLabStorageOptions,
                              StorageOptions)

__all__ = [
    "get_filesystem",
    "AbstractFileSystem",
    "StorageOptions",
    "AwsStorageOptions",
    "AzureStorageOptions",
    "GcsStorageOptions",
    "GitHubStorageOptions",
    "GitLabStorageOptions",
    "BaseStorageOptions",
]



import datetime as dt
import importlib

import msgspec

from ... import settings
from ...settings.backend import BACKEND_PROPERTIES
from ..base import BaseConfig


class JobQueueBackendConfig(BaseConfig):
    """
    Job Queue backend configuration for FlowerPower.
    Inherits from BaseConfig and adapts Redis logic.
    """

    type: str | None = msgspec.field(default=None)
    uri: str | None = msgspec.field(default=None)
    username: str | None = msgspec.field(default=None)
    password: str | None = msgspec.field(default=None)
    host: str | None = msgspec.field(default=None)
    port: int | None = msgspec.field(default=None)
    database: int | None = msgspec.field(default=None)
    ssl: bool = msgspec.field(default=False)
    cert_file: str | None = msgspec.field(default=None)
    key_file: str | None = msgspec.field(default=None)
    ca_file: str | None = msgspec.field(default=None)
    verify_ssl: bool = msgspec.field(default=False)


class APSDataStoreConfig(JobQueueBackendConfig):
    type: str = msgspec.field(default=settings.APS_BACKEND_DS)
    username: str = msgspec.field(default=settings.APS_BACKEND_DS_USERNAME)
    password: str = msgspec.field(default=settings.APS_BACKEND_DS_PASSWORD)

    host: str = msgspec.field(default=settings.APS_BACKEND_DS_HOST)
    port: int = msgspec.field(default=settings.APS_BACKEND_DS_PORT)
    database: str = msgspec.field(default=settings.APS_BACKEND_DS_DB)
    schema: str | None = msgspec.field(default=settings.APS_BACKEND_DS_SCHEMA)

    def update_from_settings(self, settings):
        if self.host == BACKEND_PROPERTIES[self.type]["default_host"]:
            self.host = settings.APS_BACKEND_DS_HOST
        if self.port == BACKEND_PROPERTIES[self.type]["default_port"]:
            self.port = settings.APS_BACKEND_DS_PORT
        if self.database == BACKEND_PROPERTIES[self.type]["default_database"]:
            self.database = settings.APS_BACKEND_DS_DB
        if self.username == BACKEND_PROPERTIES[self.type]["default_username"]:
            self.username = settings.APS_BACKEND_DS_USERNAME
        if self.password == BACKEND_PROPERTIES[self.type]["default_password"]:
            self.password = settings.APS_BACKEND_DS_PASSWORD

    def update_from_env(self):
        importlib.reload(settings.job_queue)
        importlib.reload(settings)
        self.update_from_settings(settings=settings)


class APSEventBrokerConfig(JobQueueBackendConfig):
    type: str = msgspec.field(default=settings.APS_BACKEND_EB)
    username: str = msgspec.field(default=settings.APS_BACKEND_EB_USERNAME)
    password: str = msgspec.field(default=settings.APS_BACKEND_EB_PASSWORD)
    host: str = msgspec.field(default=settings.APS_BACKEND_EB_HOST)
    port: int = msgspec.field(default=settings.APS_BACKEND_EB_PORT)
    database: str = msgspec.field(default=settings.APS_BACKEND_EB_DB)
    from_ds_sqla: bool = msgspec.field(
        default_factory=lambda: settings.APS_BACKEND_EB == "postgresql"
        and settings.APS_BACKEND_DS == "postgresql"
    )

    def update_from_settings(self, settings):
        if self.host == BACKEND_PROPERTIES[self.type]["default_host"]:
            self.host = settings.APS_BACKEND_EB_HOST
        if self.port == BACKEND_PROPERTIES[self.type]["default_port"]:
            self.port = settings.APS_BACKEND_EB_PORT
        if self.username == BACKEND_PROPERTIES[self.type]["default_username"]:
            self.username = settings.APS_BACKEND_EB_USERNAME
        if self.password == BACKEND_PROPERTIES[self.type]["default_password"]:
            self.password = settings.APS_BACKEND_EB_PASSWORD
        if self.database == BACKEND_PROPERTIES[self.type]["default_database"]:
            self.database = settings.APS_BACKEND_EB_DB

    def update_from_env(self):
        importlib.reload(settings.job_queue)
        importlib.reload(settings)
        self.update_from_settings(settings=settings)


class APSBackendConfig(BaseConfig):
    data_store: APSDataStoreConfig = msgspec.field(default_factory=APSDataStoreConfig)
    event_broker: APSEventBrokerConfig = msgspec.field(
        default_factory=APSEventBrokerConfig
    )
    cleanup_interval: int | float | dt.timedelta = msgspec.field(
        default=settings.APS_CLEANUP_INTERVAL
    )  # int in seconds
    max_concurrent_jobs: int = msgspec.field(default=settings.APS_MAX_CONCURRENT_JOBS)
    default_job_executor: str | None = msgspec.field(default=settings.EXECUTOR)
    num_workers: int | None = msgspec.field(default=settings.APS_NUM_WORKERS)

    def __post_init__(self):
        self.data_store.update_from_env()
        self.event_broker.update_from_env()


class RQBackendConfig(JobQueueBackendConfig):
    type: str = msgspec.field(default="redis")
    username: str | None = msgspec.field(default=settings.RQ_BACKEND_USERNAME)
    password: str | None = msgspec.field(default=settings.RQ_BACKEND_PASSWORD)
    host: str = msgspec.field(default=settings.RQ_BACKEND_HOST)
    port: int = msgspec.field(default=settings.RQ_BACKEND_PORT)
    database: int = msgspec.field(default=settings.RQ_BACKEND_DB)
    queues: str | list[str] = msgspec.field(default_factory=lambda: settings.RQ_QUEUES)
    num_workers: int = msgspec.field(default=settings.RQ_NUM_WORKERS)  # int in seconds

    def update_from_settings(self, settings):
        if self.host == BACKEND_PROPERTIES[self.type]["default_host"]:
            self.host = settings.RQ_BACKEND_HOST
        if self.port == BACKEND_PROPERTIES[self.type]["default_port"]:
            self.port = settings.RQ_BACKEND_PORT
        if self.database == BACKEND_PROPERTIES[self.type]["default_database"]:
            self.database = settings.RQ_BACKEND_DB
        if self.username == BACKEND_PROPERTIES[self.type]["default_username"]:
            self.username = settings.RQ_BACKEND_USERNAME
        if self.password == BACKEND_PROPERTIES[self.type]["default_password"]:
            self.password = settings.RQ_BACKEND_PASSWORD

    def update_from_env(self):
        importlib.reload(settings.job_queue)
        importlib.reload(settings)
        self.update_from_settings(settings=settings)

    def __post_init__(self):
        if isinstance(self.queues, str):
            self.queues = self.queues.replace(" ", "").split(",")
        elif not isinstance(self.queues, list):
            raise ValueError(
                f"Invalid queues type: {type(self.queues)}. Must be a string or a list."
            )
        self.update_from_env()


class HueyBackendConfig(JobQueueBackendConfig):
    pass


class JobQueueConfig(BaseConfig):
    type: str | None = msgspec.field(default="rq")
    backend: dict | None = msgspec.field(default=None)

    def __post_init__(self):
        if self.type is not None:
            self.type = self.type.lower()
            if self.type == "rq":
                if self.backend is None:
                    self.backend = RQBackendConfig()
                else:
                    if isinstance(self.backend, dict):
                        self.backend = RQBackendConfig.from_dict(self.backend)
                    elif isinstance(self.backend, RQBackendConfig):
                        pass
                    else:
                        raise ValueError(
                            f"Invalid backend type for RQ: {type(self.backend)}"
                        )
            elif self.type == "apscheduler":
                if self.backend is None:
                    self.backend = APSBackendConfig()
                else:
                    if isinstance(self.backend, dict):
                        self.backend = APSBackendConfig.from_dict(self.backend)
                    elif isinstance(self.backend, APSBackendConfig):
                        pass
                    else:
                        raise ValueError(
                            f"Invalid backend type for APScheduler: {type(self.backend)}"
                        )

            elif self.type == "huey":
                if self.backend is None:
                    self.backend = HueyBackendConfig()
                else:
                    if isinstance(self.backend, dict):
                        self.backend = HueyBackendConfig.from_dict(self.backend)
                    elif isinstance(self.backend, HueyBackendConfig):
                        pass
                    else:
                        raise ValueError(
                            f"Invalid backend type for Huey: {type(self.backend)}"
                        )
                    self.backend = HueyBackendConfig(**self.backend)
            else:
                raise ValueError(
                    f"Invalid job queue type: {self.type}. Valid types: {['rq', 'apscheduler', 'huey']}"
                )

    def update_type(self, type: str):
        if type != self.type:
            self.type = type
            self.backend = None
            self.__post_init__()



import datetime as dt
import re
from functools import lru_cache
from typing import Any
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

import pyarrow as pa
import pyarrow.compute as pc
from sqlglot import exp, parse_one

from .polars import pl


@lru_cache(maxsize=128)
def timestamp_from_string(
    timestamp_str: str,
    tz: str | None = None,
    naive: bool = False,
) -> dt.datetime | dt.date | dt.time:
    """
    Converts a timestamp string (ISO 8601 format) into a datetime, date, or time object
    using only standard Python libraries.

    Handles strings with or without timezone information (e.g., '2023-01-01T10:00:00+02:00',
    '2023-01-01', '10:00:00'). Supports timezone offsets like '+HH:MM' or '+HHMM'.
    For named timezones (e.g., 'Europe/Paris'), requires Python 3.9+ and the 'tzdata'
    package to be installed.

    Args:
        timestamp_str (str): The string representation of the timestamp (ISO 8601 format).
        tz (str, optional): Target timezone identifier (e.g., 'UTC', '+02:00', 'Europe/Paris').
            If provided, the output datetime/time will be localized or converted to this timezone.
            Defaults to None.
        naive (bool, optional): If True, return a naive datetime/time (no timezone info),
            even if the input string or `tz` parameter specifies one. Defaults to False.

    Returns:
        Union[dt.datetime, dt.date, dt.time]: The parsed datetime, date, or time object.

    Raises:
        ValueError: If the timestamp string format is invalid or the timezone is
                    invalid/unsupported.
    """

    # Regex to parse timezone offsets like +HH:MM or +HHMM
    _TZ_OFFSET_REGEX = re.compile(r"([+-])(\d{2}):?(\d{2})")

    def _parse_tz_offset(tz_str: str) -> dt.tzinfo | None:
        """Parses a timezone offset string into a timezone object."""
        match = _TZ_OFFSET_REGEX.fullmatch(tz_str)
        if match:
            sign, hours, minutes = match.groups()
            offset_seconds = (int(hours) * 3600 + int(minutes) * 60) * (
                -1 if sign == "-" else 1
            )
            if abs(offset_seconds) >= 24 * 3600:
                raise ValueError(f"Invalid timezone offset: {tz_str}")
            return dt.timezone(dt.timedelta(seconds=offset_seconds), name=tz_str)
        return None

    def _get_tzinfo(tz_identifier: str | None) -> dt.tzinfo | None:
        """Gets a tzinfo object from a string (offset or IANA name)."""
        if tz_identifier is None:
            return None
        if tz_identifier.upper() == "UTC":
            return dt.timezone.utc

        # Try parsing as offset first
        offset_tz = _parse_tz_offset(tz_identifier)
        if offset_tz:
            return offset_tz

        # Try parsing as IANA name using zoneinfo (if available)
        if ZoneInfo:
            try:
                return ZoneInfo(tz_identifier)
            except ZoneInfoNotFoundError:
                raise ValueError(
                    f"Timezone '{tz_identifier}' not found. Install 'tzdata' or use offset format."
                )
            except Exception as e:  # Catch other potential zoneinfo errors
                raise ValueError(f"Error loading timezone '{tz_identifier}': {e}")
        else:
            # zoneinfo not available
            raise ValueError(
                f"Invalid timezone: '{tz_identifier}'. Use offset format (e.g., '+02:00') "
                "or run Python 3.9+ with 'tzdata' installed for named timezones."
            )

    target_tz: dt.tzinfo | None = _get_tzinfo(tz)
    parsed_obj: dt.datetime | dt.date | dt.time | None = None

    # Preprocess: Replace space separator, strip whitespace
    processed_str = timestamp_str.strip().replace(" ", "T")

    # Attempt parsing (datetime, date, time) using fromisoformat
    try:
        # Python < 3.11 fromisoformat has limitations (e.g., no Z, no +HHMM offset)
        # This implementation assumes Python 3.11+ for full ISO 8601 support via fromisoformat
        # or that input strings use formats compatible with older versions (e.g., +HH:MM)
        parsed_obj = dt.datetime.fromisoformat(processed_str)
    except ValueError:
        try:
            parsed_obj = dt.date.fromisoformat(processed_str)
        except ValueError:
            try:
                # Time parsing needs care, especially with offsets in older Python
                parsed_obj = dt.time.fromisoformat(processed_str)
            except ValueError:
                # Add fallback for simple HH:MM:SS if needed, though less robust
                # try:
                #     parsed_obj = dt.datetime.strptime(processed_str, "%H:%M:%S").time()
                # except ValueError:
                raise ValueError(f"Invalid timestamp format: '{timestamp_str}'")

    # Apply timezone logic if we have a datetime or time object
    if isinstance(parsed_obj, (dt.datetime, dt.time)):
        is_aware = (
            parsed_obj.tzinfo is not None
            and parsed_obj.tzinfo.utcoffset(
                parsed_obj if isinstance(parsed_obj, dt.datetime) else None
            )
            is not None
        )

        if target_tz:
            if is_aware:
                # Convert existing aware object to target timezone (only for datetime)
                if isinstance(parsed_obj, dt.datetime):
                    parsed_obj = parsed_obj.astimezone(target_tz)
                # else: dt.time cannot be converted without a date context. Keep original tz.
            else:
                # Localize naive object to target timezone
                parsed_obj = parsed_obj.replace(tzinfo=target_tz)
            is_aware = True  # Object is now considered aware

        # Handle naive flag: remove tzinfo if requested
        if naive and is_aware:
            parsed_obj = parsed_obj.replace(tzinfo=None)

    # If it's a date object, tz/naive flags are ignored
    elif isinstance(parsed_obj, dt.date):
        pass

    return parsed_obj


# Compile regex patterns once for efficiency
SPLIT_PATTERN = re.compile(
    r"<=|<|>=|>|=|!=|\s+[n,N][o,O][t,T]\s+[i,I][n,N]\s+|\s+[i,I][n,N]\s+|"
    r"\s+[i,I][s,S]\s+[n,N][o,O][t,T]\s+[n,N][u,U][l,L]{2}\s+|\s+[i,I][s,S]\s+[n,N][u,U][l,L]{2}\s+"
)
LOGICAL_OPERATORS_PATTERN = re.compile(
    r"\s+[a,A][n,N][d,D] [n,N][o,O][t,T]\s+|\s+[a,A][n,N][d,D]\s+|"
    r"\s+[o,O][r,R] [n,N][o,O][t,T]\s+|\s+[o,O][r,R]\s+"
)


def sql2pyarrow_filter(string: str, schema: pa.Schema) -> pc.Expression:
    """
    Generates a filter expression for PyArrow based on a given string and schema.

    Parameters:
        string (str): The string containing the filter expression.
        schema (pa.Schema): The PyArrow schema used to validate the filter expression.

    Returns:
        pc.Expression: The generated filter expression.

    Raises:
        ValueError: If the input string is invalid or contains unsupported operations.
    """

    def parse_value(val: str, type_: pa.DataType) -> Any:
        """Parse and convert value based on the field type."""
        if isinstance(val, (tuple, list)):
            return type(val)(parse_value(v, type_) for v in val)

        if pa.types.is_timestamp(type_):
            return timestamp_from_string(val, exact=False, tz=type_.tz)
        elif pa.types.is_date(type_):
            return timestamp_from_string(val, exact=True).date()
        elif pa.types.is_time(type_):
            return timestamp_from_string(val, exact=True).time()

        elif pa.types.is_integer(type_):
            return int(float(val.strip("'").replace(",", ".")))
        elif pa.types.is_floating(type_):
            return float(val.strip("'").replace(",", "."))
        elif pa.types.is_boolean(type_):
            return val.lower().strip("'") in ("true", "1", "yes")
        else:
            return val.strip("'")

    def _parse_part(part: str) -> pc.Expression:
        match = SPLIT_PATTERN.search(part)
        if not match:
            raise ValueError(f"Invalid condition: {part}")

        sign = match.group().lower().strip()
        field, val = [p.strip() for p in SPLIT_PATTERN.split(part)]

        if field not in schema.names:
            raise ValueError(f"Unknown field: {field}")

        type_ = schema.field(field).type
        val = parse_value(val, type_)

        operations = {
            ">=": lambda f, v: pc.field(f) >= v,
            ">": lambda f, v: pc.field(f) > v,
            "<=": lambda f, v: pc.field(f) <= v,
            "<": lambda f, v: pc.field(f) < v,
            "=": lambda f, v: pc.field(f) == v,
            "!=": lambda f, v: pc.field(f) != v,
            "in": lambda f, v: pc.field(f).isin(v),
            "not in": lambda f, v: ~pc.field(f).isin(v),
            "is null": lambda f, v: pc.field(f).is_null(nan_is_null=True),
            "is not null": lambda f, v: ~pc.field(f).is_null(nan_is_null=True),
        }

        if sign not in operations:
            raise ValueError(f"Unsupported operation: {sign}")

        return operations[sign](field, val)

    parts = LOGICAL_OPERATORS_PATTERN.split(string)
    operators = [op.lower().strip() for op in LOGICAL_OPERATORS_PATTERN.findall(string)]

    if len(parts) == 1:
        return _parse_part(parts[0])

    expr = _parse_part(parts[0])
    for part, operator in zip(parts[1:], operators):
        if operator == "and":
            expr = expr & _parse_part(part)
        elif operator == "and not":
            expr = expr & ~_parse_part(part)
        elif operator == "or":
            expr = expr | _parse_part(part)
        elif operator == "or not":
            expr = expr | ~_parse_part(part)
        else:
            raise ValueError(f"Unsupported logical operator: {operator}")

    return expr


def sql2polars_filter(string: str, schema: pl.Schema) -> pl.Expr:
    """
    Generates a filter expression for Polars based on a given string and schema.

    Parameters:
        string (str): The string containing the filter expression.
        schema (pl.Schema): The Polars schema used to validate the filter expression.

    Returns:
        pl.Expr: The generated filter expression.

    Raises:
        ValueError: If the input string is invalid or contains unsupported operations.
    """

    def parse_value(val: str, dtype: pl.DataType) -> Any:
        """Parse and convert value based on the field type."""
        if isinstance(val, (tuple, list)):
            return type(val)(parse_value(v, dtype) for v in val)

        if dtype == pl.Datetime:
            return timestamp_from_string(val, exact=False, tz=dtype.time_zone)
        elif dtype == pl.Date:
            return timestamp_from_string(val, exact=True).date()
        elif dtype == pl.Time:
            return timestamp_from_string(val, exact=True).time()
        elif dtype in (pl.Int8, pl.Int16, pl.Int32, pl.Int64):
            return int(float(val.strip("'").replace(",", ".")))
        elif dtype in (pl.Float32, pl.Float64):
            return float(val.strip("'").replace(",", "."))
        elif dtype == pl.Boolean:
            return val.lower().strip("'") in ("true", "1", "yes")
        else:
            return val.strip("'")

    def _parse_part(part: str) -> pl.Expr:
        match = SPLIT_PATTERN.search(part)
        if not match:
            raise ValueError(f"Invalid condition: {part}")

        sign = match.group().lower().strip()
        field, val = [p.strip() for p in SPLIT_PATTERN.split(part)]

        if field not in schema.names():
            raise ValueError(f"Unknown field: {field}")

        dtype = schema[field]
        val = parse_value(val, dtype)

        operations = {
            ">=": lambda f, v: pl.col(f) >= v,
            ">": lambda f, v: pl.col(f) > v,
            "<=": lambda f, v: pl.col(f) <= v,
            "<": lambda f, v: pl.col(f) < v,
            "=": lambda f, v: pl.col(f) == v,
            "!=": lambda f, v: pl.col(f) != v,
            "in": lambda f, v: pl.col(f).is_in(v),
            "not in": lambda f, v: ~pl.col(f).is_in(v),
            "is null": lambda f, v: pl.col(f).is_null(),
            "is not null": lambda f, v: pl.col(f).is_not_null(),
        }

        if sign not in operations:
            raise ValueError(f"Unsupported operation: {sign}")

        return operations[sign](field, val)

    parts = LOGICAL_OPERATORS_PATTERN.split(string)
    operators = [op.lower().strip() for op in LOGICAL_OPERATORS_PATTERN.findall(string)]

    if len(parts) == 1:
        return _parse_part(parts[0])

    expr = _parse_part(parts[0])
    for part, operator in zip(parts[1:], operators):
        if operator == "and":
            expr = expr & _parse_part(part)
        elif operator == "and not":
            expr = expr & ~_parse_part(part)
        elif operator == "or":
            expr = expr | _parse_part(part)
        elif operator == "or not":
            expr = expr | ~_parse_part(part)
        else:
            raise ValueError(f"Unsupported logical operator: {operator}")

    return expr


def get_table_names(sql_query):
    return [table.name for table in parse_one(sql_query).find_all(exp.Table)]



import os

from .backend import BACKEND_PROPERTIES
from .executor import EXECUTOR, EXECUTOR_MAX_WORKERS, EXECUTOR_NUM_CPUS

# WORKER
JOB_QUEUE_TYPE = os.getenv("FP_JOB_QUEUE_TYPE", "rq")

# RQ WORKER
RQ_BACKEND = os.getenv("FP_RQ_BACKEND", "redis")
RQ_BACKEND_HOST = os.getenv(
    "FP_RQ_BACKEND_HOST", BACKEND_PROPERTIES[RQ_BACKEND]["default_host"]
)
RQ_BACKEND_PORT = int(
    os.getenv("FP_RQ_BACKEND_PORT", BACKEND_PROPERTIES[RQ_BACKEND]["default_port"])
)
RQ_BACKEND_DB = int(
    os.getenv("FP_RQ_BACKEND_DB", BACKEND_PROPERTIES[RQ_BACKEND]["default_database"])
)
RQ_BACKEND_PASSWORD = os.getenv(
    "FP_RQ_BACKEND_PASSWORD", BACKEND_PROPERTIES[RQ_BACKEND]["default_password"]
)
RQ_BACKEND_USERNAME = os.getenv(
    "FP_RQ_BACKEND_USERNAME", BACKEND_PROPERTIES[RQ_BACKEND]["default_username"]
)
RQ_QUEUES = (
    os.getenv("FP_RQ_QUEUES", "default, high, low, scheduler")
    .replace(" ", "")
    .split(",")
)
RQ_NUM_WORKERS = int(os.getenv("FP_RQ_NUM_WORKERS", EXECUTOR_NUM_CPUS))

# APS WORKER
APS_BACKEND_DS = os.getenv("FP_APS_BACKEND_DS", "postgresql")
APS_BACKEND_DS_HOST = os.getenv(
    "FP_APS_BACKEND_DS_HOST", BACKEND_PROPERTIES[APS_BACKEND_DS]["default_host"]
)
APS_BACKEND_DS_PORT = int(
    os.getenv(
        "FP_APS_BACKEND_DS_PORT", BACKEND_PROPERTIES[APS_BACKEND_DS]["default_port"]
    )
)
APS_BACKEND_DS_DB = os.getenv(
    "FP_APS_BACKEND_DS_DB", BACKEND_PROPERTIES[APS_BACKEND_DS]["default_database"]
)
APS_BACKEND_DS_USERNAME = os.getenv(
    "FP_APS_BACKEND_DS_USERNAME", BACKEND_PROPERTIES[APS_BACKEND_DS]["default_username"]
)
APS_BACKEND_DS_PASSWORD = os.getenv(
    "FP_APS_BACKEND_DS_PASSWORD", BACKEND_PROPERTIES[APS_BACKEND_DS]["default_password"]
)
APS_BACKEND_DS_SCHEMA = os.getenv("FP_APS_BACKEND_DS_SCHEMA", "flowerpower")

APS_BACKEND_EB = os.getenv("FP_APS_BACKEND_EB", "postgresql")
APS_BACKEND_EB_HOST = os.getenv(
    "FP_APS_BACKEND_EB_HOST", BACKEND_PROPERTIES[APS_BACKEND_EB]["default_host"]
)
APS_BACKEND_EB_PORT = int(
    os.getenv(
        "FP_APS_BACKEND_EB_PORT", BACKEND_PROPERTIES[APS_BACKEND_EB]["default_port"]
    )
)
APS_BACKEND_EB_DB = os.getenv(
    "FP_APS_BACKEND_EB_DB", BACKEND_PROPERTIES[APS_BACKEND_EB]["default_database"]
)
APS_BACKEND_EB_USERNAME = os.getenv(
    "FP_APS_BACKEND_EB_USERNAME", BACKEND_PROPERTIES[APS_BACKEND_EB]["default_username"]
)
APS_BACKEND_EB_PASSWORD = os.getenv(
    "FP_APS_BACKEND_EB_PASSWORD", BACKEND_PROPERTIES[APS_BACKEND_EB]["default_password"]
)

APS_CLEANUP_INTERVAL = int(os.getenv("FP_APS_CLEANUP_INTERVAL", 300))
APS_MAX_CONCURRENT_JOBS = int(os.getenv("FP_APS_MAX_CONCURRENT_JOBS", 10))
APS_DEFAULT_EXECUTOR = os.getenv("FP_APS_DEFAULT_EXECUTOR", EXECUTOR)
APS_NUM_WORKERS = int(os.getenv("FP_APS_NUM_WORKERS", EXECUTOR_MAX_WORKERS))



import importlib
import sys

from dill import dumps, loads


def patch_pickle():
    """
    Patch the pickle serializer in the apscheduler module.

    This function replaces the `dumps` and `loads` functions in the `apscheduler.serializers.pickle` module
    with custom implementations.

    This is useful when you want to modify the behavior of the pickle serializer used by the apscheduler module.

    Example usage:
    patch_pickle()

    """
    sys.modules["apscheduler.serializers.pickle"].dumps = dumps
    sys.modules["apscheduler.serializers.pickle"].loads = loads


if importlib.util.find_spec("apscheduler"):
    from apscheduler._structures import Job, Schedule, Task

    def job_to_dict(job):
        return {
            "id": str(job.id),
            "task_id": job.task_id,
            "args": [str(arg) for arg in job.args],
            "kwargs": job.kwargs,
            "schedule_id": job.schedule_id,
            "scheduled_fire_time": (
                job.scheduled_fire_time.isoformat() if job.scheduled_fire_time else None
            ),
            "jitter": job.jitter.total_seconds(),
            "start_deadline": (
                job.start_deadline.isoformat() if job.start_deadline else None
            ),
            "result_expiration_time": job.result_expiration_time.total_seconds(),
            "created_at": job.created_at.isoformat(),
            "acquired_by": job.acquired_by,
            "acquired_until": (
                job.acquired_until.isoformat() if job.acquired_until else None
            ),
        }

    Job.to_dict = job_to_dict

    def task_to_dict(task):
        return {
            "id": task.id,
            "func": task.func,
            "job_executor": task.job_executor,
            "max_running_jobs": task.max_running_jobs,
            "misfire_grace_time": task.misfire_grace_time,
        }

    Task.to_dict = task_to_dict

    def schedule_to_dict(schedule):
        return {
            "id": schedule.id,
            "task_id": schedule.task_id,
            "trigger": str(schedule.trigger),
            "args": [str(arg) for arg in schedule.args],
            "kwargs": schedule.kwargs,
            "paused": schedule.paused,
            "coalesce": schedule.coalesce.name if schedule.coalesce else None,
            "misfire_grace_time": schedule.misfire_grace_time,
            "max_jitter": schedule.max_jitter,
            "next_fire_time": (
                schedule.next_fire_time.isoformat() if schedule.next_fire_time else None
            ),
            "last_fire_time": (
                schedule.last_fire_time.isoformat() if schedule.last_fire_time else None
            ),
            "acquired_by": schedule.acquired_by,
            "acquired_until": (
                schedule.acquired_until.isoformat() if schedule.acquired_until else None
            ),
        }

    Schedule.to_dict = schedule_to_dict



import msgspec

from ... import settings
from ...fs import AbstractFileSystem, BaseStorageOptions, get_filesystem
from ..base import BaseConfig
from .adapter import AdapterConfig
from .job_queue import JobQueueConfig


class ProjectConfig(BaseConfig):
    """A configuration class for managing project-level settings in FlowerPower.

    This class handles project-wide configuration including job queue and adapter settings.
    It supports loading from and saving to YAML files, with filesystem abstraction.

    Attributes:
        name (str | None): The name of the project.
        job_queue (JobQueueConfig): Configuration for the job queue component.
        adapter (AdapterConfig): Configuration for the adapter component.

    Example:
        ```python
        # Create a new project config
        project = ProjectConfig(name="my-project")

        # Load existing project config
        project = ProjectConfig.load(base_dir="path/to/project")

        # Save project config
        project.save(base_dir="path/to/project")
        ```
    """

    name: str | None = msgspec.field(default=None)
    job_queue: JobQueueConfig = msgspec.field(default_factory=JobQueueConfig)
    adapter: AdapterConfig = msgspec.field(default_factory=AdapterConfig)

    def __post_init__(self):
        if isinstance(self.job_queue, dict):
            self.job_queue = JobQueueConfig.from_dict(self.job_queue)
        if isinstance(self.adapter, dict):
            self.adapter = AdapterConfig.from_dict(self.adapter)

    @classmethod
    def load(
        cls,
        base_dir: str = ".",
        name: str | None = None,
        job_queue_type: str | None = None,
        fs: AbstractFileSystem | None = None,
        storage_options: dict | BaseStorageOptions | None = {},
    ):
        """Load project configuration from a YAML file.

        Args:
            base_dir (str, optional): Base directory for the project. Defaults to ".".
            name (str | None, optional): Project name. Defaults to None.
            job_queue_type (str | None, optional): Type of job queue to use. Defaults to None.
            fs (AbstractFileSystem | None, optional): Filesystem to use. Defaults to None.
            storage_options (dict | Munch, optional): Options for filesystem. Defaults to empty Munch.

        Returns:
            ProjectConfig: Loaded project configuration.

        Example:
            ```python
            project = ProjectConfig.load(
                base_dir="my_project",
                name="pipeline1",
                job_queue_type="rq"
                )
            ```
        """
        if fs is None:
            fs = get_filesystem(
                base_dir, cached=False, dirfs=True, storage_options=storage_options
            )
        if fs.exists("conf/project.yml"):
            project = ProjectConfig.from_yaml(path="conf/project.yml", fs=fs)
        else:
            project = ProjectConfig(name=name)
        if job_queue_type is not None:
            if job_queue_type != project.job_queue.type:
                project.job_queue.update_type(job_queue_type)

        return project

    def save(
        self,
        base_dir: str = ".",
        fs: AbstractFileSystem | None = None,
        storage_options: dict | BaseStorageOptions | None = {},
    ):
        """Save project configuration to a YAML file.

        Args:
            base_dir (str, optional): Base directory for the project. Defaults to ".".
            fs (AbstractFileSystem | None, optional): Filesystem to use. Defaults to None.
            storage_options (dict | Munch, optional): Options for filesystem. Defaults to empty Munch.

        Example:
            ```python
            project_config.save(base_dir="my_project")
            ```
        """
        if fs is None:
            fs = get_filesystem(
                base_dir, cached=True, dirfs=True, storage_options=storage_options
            )

        fs.makedirs("conf", exist_ok=True)
        self.to_yaml(path="conf/project.yml", fs=fs)


def init_project_config(
    base_dir: str = ".",
    name: str | None = None,
    job_queue_type: str | None = None,
    fs: AbstractFileSystem | None = None,
    storage_options: dict | BaseStorageOptions | None = {},
):
    """Initialize a new project configuration.

    This function creates a new project configuration and saves it to disk.

    Args:
        base_dir (str, optional): Base directory for the project. Defaults to ".".
        name (str | None, optional): Project name. Defaults to None.
        job_queue_type (str | None, optional): Type of job queue to use. Defaults to None.
        fs (AbstractFileSystem | None, optional): Filesystem to use. Defaults to None.
        storage_options (dict | Munch, optional): Options for filesystem. Defaults to empty Munch.

    Returns:
        ProjectConfig: The initialized project configuration.

    Example:
        ```python
        project = init_project_config(
            base_dir="my_project",
            name="test_project",
            job_queue_type="rq"
        )
        ```
    """
    project = ProjectConfig.load(
        base_dir=base_dir,
        name=name,
        job_queue_type=job_queue_type,
        fs=fs,
        storage_options=storage_options,
    )
    project.save(base_dir=base_dir, fs=fs, storage_options=storage_options)
    return project



import datetime as dt
from enum import Enum
from typing import Any, Dict, Type

from apscheduler.triggers.calendarinterval import CalendarIntervalTrigger
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.date import DateTrigger
from apscheduler.triggers.interval import IntervalTrigger

from ..base import BaseTrigger


class TriggerType(Enum):
    CRON = "cron"
    INTERVAL = "interval"
    CALENDARINTERVAL = "calendarinterval"
    DATE = "date"


# Mapping of trigger type to its class and allowed kwargs
TRIGGER_CONFIG: Dict[TriggerType, Dict[str, Any]] = {
    TriggerType.CRON: {
        "class": CronTrigger,
        "kwargs": [
            "crontab",
            "year",
            "month",
            "week",
            "day",
            "day_of_week",
            "hour",
            "minute",
            "second",
            "start_time",
            "end_time",
            "timezone",
        ],
    },
    TriggerType.INTERVAL: {
        "class": IntervalTrigger,
        "kwargs": [
            "weeks",
            "days",
            "hours",
            "minutes",
            "seconds",
            "microseconds",
            "start_time",
            "end_time",
        ],
    },
    TriggerType.CALENDARINTERVAL: {
        "class": CalendarIntervalTrigger,
        "kwargs": [
            "years",
            "months",
            "weeks",
            "days",
            "hour",
            "minute",
            "second",
            "start_date",
            "end_date",
            "timezone",
        ],
    },
    TriggerType.DATE: {
        "class": DateTrigger,
        "kwargs": [
            "run_time",
        ],
    },
}


class APSTrigger(BaseTrigger):
    """
    Implementation of BaseTrigger for APScheduler.

    Provides a factory for creating APScheduler trigger instances
    with validation and filtering of keyword arguments.
    """

    trigger_type: TriggerType

    def __init__(self, trigger_type: str):
        """
        Initialize APSchedulerTrigger with a trigger type.

        Args:
            trigger_type (str): The type of trigger (cron, interval, calendarinterval, date).

        Raises:
            ValueError: If the trigger_type is invalid.
        """
        try:
            self.trigger_type = TriggerType(trigger_type.lower())
        except ValueError:
            valid_types = [t.value for t in TriggerType]
            raise ValueError(
                f"Invalid trigger type '{trigger_type}'. Valid types are: {valid_types}"
            )

    def _get_allowed_kwargs(self) -> set:
        """Return the set of allowed kwargs for the current trigger type."""
        return set(TRIGGER_CONFIG[self.trigger_type]["kwargs"])

    def _check_kwargs(self, **kwargs) -> None:
        """
        Validate that all provided kwargs are allowed for the trigger type.

        Raises:
            ValueError: If any kwarg is not allowed.
        """
        allowed = self._get_allowed_kwargs()
        invalid = [k for k in kwargs if k not in allowed]
        if invalid:
            raise ValueError(
                f"Invalid argument(s) for trigger type '{self.trigger_type.value}': {invalid}. "
                f"Allowed arguments are: {sorted(allowed)}"
            )

    def _filter_kwargs(self, **kwargs) -> Dict[str, Any]:
        """
        Filter kwargs to only those allowed for the trigger type and not None.

        Returns:
            Dict[str, Any]: Filtered kwargs.
        """
        allowed = self._get_allowed_kwargs()
        return {k: v for k, v in kwargs.items() if k in allowed and v is not None}

    def get_trigger_instance(self, **kwargs) -> Any:
        """
        Create and return an APScheduler trigger instance based on the trigger type.

        Args:
            **kwargs: Keyword arguments for the trigger.

        Returns:
            Any: An APScheduler trigger instance.

        Raises:
            ValueError: If invalid arguments are provided or trigger type is unknown.
        """
        self._check_kwargs(**kwargs)
        filtered_kwargs = self._filter_kwargs(**kwargs)
        trigger_cls: Type = TRIGGER_CONFIG[self.trigger_type]["class"]

        if self.trigger_type == TriggerType.CRON:
            crontab = filtered_kwargs.pop("crontab", None)
            if crontab:
                return trigger_cls.from_crontab(crontab)
            return trigger_cls(**filtered_kwargs)
        elif self.trigger_type == TriggerType.INTERVAL:
            return trigger_cls(**filtered_kwargs)
        elif self.trigger_type == TriggerType.CALENDARINTERVAL:
            return trigger_cls(**filtered_kwargs)
        elif self.trigger_type == TriggerType.DATE:
            # Default to now if not specified
            if "run_time" not in filtered_kwargs:
                filtered_kwargs["run_time"] = dt.datetime.now()
            return trigger_cls(**filtered_kwargs)
        else:
            # This should never be reached due to Enum validation in __init__
            raise ValueError(f"Unknown trigger type: {self.trigger_type.value}")


# End of file



# -*- coding: utf-8 -*-
# pylint: disable=logging-fstring-interpolation
# flake8: noqa: E501
"""Pipeline Job Queue."""

import datetime as dt
from typing import Any, Callable
from uuid import UUID

from loguru import logger
from rich import print as rprint

from .. import settings
# Import necessary config types
from ..cfg import PipelineConfig, ProjectConfig
from ..fs import AbstractFileSystem
from ..job_queue import JobQueueManager
from ..utils.logging import setup_logging
from .registry import PipelineRegistry

setup_logging()


class PipelineJobQueue:
    """Handles scheduling of pipeline runs via a configured job queue backend."""

    def __init__(
        self,
        project_cfg: ProjectConfig,
        fs: AbstractFileSystem,
        cfg_dir: str,
        pipelines_dir: str,
        job_queue_type: str | None = None,
    ):
        """Initialize PipelineJobQueue.

        Args:
            project_cfg: The project configuration object.
            fs: The file system to use for file operations.
            cfg_dir: The directory for configuration files.
            pipelines_dir: The directory for pipeline files.
            job_queue_type: The type of job queue to use (e.g., 'rq', 'apscheduler'). If None, defaults to the project config.
        """
        self.project_cfg = project_cfg
        self._fs = fs
        self._cfg_dir = cfg_dir
        self._pipelines_dir = pipelines_dir
        self._job_queue_type = job_queue_type or project_cfg.job_queue.type
        if not self._job_queue_type:
            # Fallback or default if not specified in project config
            self._job_queue_type = settings.JOB_QUEUE_TYPE
            logger.warning(
                f"Job queue type not specified in project config, defaulting to '{self._job_queue_type}'"
            )

    @property
    def job_queue(self):
        """
        Lazily instantiate and cache a Job queue instance.
        """
        # Lazily instantiate job queue using project_cfg attributes
        logger.debug(
            f"Instantiating job queue of type: {self._job_queue_type} for project '{self.project_cfg.name}'"
        )
        # Pass the necessary parts of project_cfg to the Job queue
        return JobQueueManager(
            type=self._job_queue_type,
            fs=self._fs,
        )

    def _get_schedule_ids(self) -> list[Any]:
        """Get all schedules from the job queue backend."""

        with self.job_queue as job_queue:
            logger.debug("Fetching schedules ids from job queue")
            return job_queue.schedule_ids

    def run_job(
        self,
        run_func: Callable,
        name: str,  # name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        config: dict | None = None,
        cache: bool | dict = False,
        executor_cfg: str | dict | Any | None = None,
        with_adapter_cfg: dict | Any | None = None,
        pipeline_adapter_cfg: dict | Any | None = None,
        project_adapter_cfg: dict | Any | None = None,
        adapter: dict[str, Any] | None = None,
        reload: bool = False,
        log_level: str | None = None,
        max_retries: int | None = None,
        retry_delay: float | None = None,
        jitter_factor: float | None = None,
        retry_exceptions: tuple | None = None,
        **kwargs,
    ) -> dict[str, Any]:
        """
        Add a job to run the pipeline immediately via the job queue queue.

        Args:
            run_func (Callable): The function to execute in the job queue (e.g., a configured PipelineRunner.run).
            name (str): The name of the pipeline (used for logging).
            inputs (dict | None): Inputs for the pipeline run.
            final_vars (list | None): Final variables for the pipeline run.
            config (dict | None): Hamilton driver config.
            cache (bool | dict): Cache configuration.
            executor_cfg (str | dict | ExecutorConfig | None): Executor configuration.
            with_adapter_cfg (dict | WithAdapterConfig | None): Adapter configuration.
            pipeline_adapter_cfg (dict | PipelineAdapterConfig | None): Pipeline adapter configuration.
            project_adapter_cfg (dict | ProjectAdapterConfig | None): Project adapter configuration.
            adapter (dict[str, Any] | None): Additional adapter configuration.
            reload (bool): Whether to reload the pipeline module.
            log_level (str | None): Log level for the run.
            max_retries (int): Maximum number of retries for the job.
            retry_delay (float): Delay between retries.
            jitter_factor (float): Jitter factor for retry delay.
            retry_exceptions (tuple): Exceptions that should trigger a retry.
            **kwargs: Additional keyword arguments passed directly to the job queue's add_job method.

        Returns:
            dict[str, Any]: The result of the job execution.
        """
        logger.debug(f"Adding immediate job for pipeline: {name}")

        pipeline_run_args = {
            # 'name' is not passed to run_func, it's part of the context already in PipelineRunner
            "inputs": inputs,
            "final_vars": final_vars,
            "config": config,
            "cache": cache,
            "executor_cfg": executor_cfg,
            "with_adapter_cfg": with_adapter_cfg,
            "pipeline_adapter_cfg": pipeline_adapter_cfg,
            "project_adapter_cfg": project_adapter_cfg,
            "adapter": adapter,
            "reload": reload,
            "log_level": log_level,
            "max_retries": max_retries,
            "retry_delay": retry_delay,
            "jitter_factor": jitter_factor,
            "retry_exceptions": retry_exceptions,
        }
        pipeline_run_args = {
            k: v for k, v in pipeline_run_args.items() if v is not None
        }
        logger.debug(
            f"Resolved arguments for target run_func for job '{name}': {pipeline_run_args}"
        )

        with self.job_queue as job_queue:
            res = job_queue.run_job(
                func=run_func,
                func_kwargs=pipeline_run_args,
                **kwargs,
            )

        return res

    def add_job(
        self,
        run_func: Callable,  # The actual function to run (e.g., PipelineRunner(...).run)
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        config: dict | None = None,
        cache: bool | dict = False,
        executor_cfg: str | dict | Any | None = None,
        with_adapter_cfg: dict | Any | None = None,
        pipeline_adapter_cfg: dict | Any | None = None,
        project_adapter_cfg: dict | Any | None = None,
        adapter: dict[str, Any] | None = None,
        result_ttl: int | dt.timedelta = 120,
        run_at: dt.datetime | None = None,
        run_in: float | dt.timedelta | None = None,
        reload: bool = False,
        log_level: str | None = None,
        max_retries: int | None = None,
        retry_delay: float | None = None,
        jitter_factor: float | None = None,
        retry_exceptions: tuple | list | None = None,
        **kwargs,  # Allow other job queue-specific args if needed
    ) -> Any:
        """
        Add a job to run the pipeline immediately via the job queue, storing the result.

        Executes the job immediately and returns the job id (UUID). The job result will be stored
        by the job queue backend for the given `result_ttl` and can be fetched using the job id.

        Args:
            run_func (Callable): The function to execute in the job queue (e.g., a configured PipelineRunner.run).
            name (str): The name of the pipeline (used for logging).
            inputs (dict | None): Inputs for the pipeline run.
            final_vars (list | None): Final variables for the pipeline run.
            config (dict | None): Hamilton driver config.
            cache (bool | dict): Cache configuration.
            executor_cfg (str | dict | ExecutorConfig | None): Executor configuration.
            with_adapter_cfg (dict | WithAdapterConfig | None): Adapter configuration.
            pipeline_adapter_cfg (dict | PipelineAdapterConfig | None): Pipeline adapter configuration.
            project_adapter_cfg (dict | ProjectAdapterConfig | None): Project adapter configuration.
            adapter (dict[str, Any] | None): Additional adapter configuration.
            reload (bool): Whether to reload the pipeline module.
            log_level (str | None): Log level for the run.
            result_ttl (int | dt.timedelta): How long the job result should be stored. Defaults to 0 (don't store).
            run_at (dt.datetime | None): Optional datetime to run the job at.
            run_in (float | dt.timedelta | None): Optional delay before running the job.
            max_retries (int): Maximum number of retries for the job.
            retry_delay (float): Delay between retries.
            jitter_factor (float): Jitter factor for retry delay.
            retry_exceptions (tuple): Exceptions that should trigger a retry.
            **kwargs: Additional keyword arguments passed directly to the job queue's add_job method.

        Returns:
            Any: The ID of the added job or the job object itself, depending on the job queue backend.
        """
        logger.debug(f"Adding immediate job with result TTL for pipeline: {name}")

        pipeline_run_args = {
            "inputs": inputs,
            "final_vars": final_vars,
            "config": config,
            "cache": cache,
            "executor_cfg": executor_cfg,
            "with_adapter_cfg": with_adapter_cfg,
            "pipeline_adapter_cfg": pipeline_adapter_cfg,
            "project_adapter_cfg": project_adapter_cfg,
            "adapter": adapter,
            "reload": reload,
            "log_level": log_level,
            "max_retries": max_retries,
            "retry_delay": retry_delay,
            "jitter_factor": jitter_factor,
            "retry_exceptions": retry_exceptions,
        }
        pipeline_run_args = {
            k: v for k, v in pipeline_run_args.items() if v is not None
        }
        logger.debug(
            f"Resolved arguments for target run_func for job (TTL) '{name}': {pipeline_run_args}"
        )

        with self.job_queue as job_queue:
            job = job_queue.add_job(
                func=run_func,
                func_kwargs=pipeline_run_args,
                result_ttl=result_ttl,
                run_at=run_at,
                run_in=run_in,
                **kwargs,
            )
        rprint(
            f"✅ Successfully added job for "
            f"[blue]{self.project_cfg.name}.{name}[/blue] with ID [green]{job if isinstance(job, (str, UUID)) else job.id}[/green]"
            f" and result TTL of {result_ttl} seconds."
        )
        return job

    # --- End Moved from PipelineManager ---

    def schedule(
        self,
        run_func: Callable,
        pipeline_cfg: PipelineConfig,
        # --- Run Parameters (passed to run_func) ---
        inputs: dict | None = None,
        final_vars: list | None = None,
        config: dict | None = None,  # Driver config
        cache: bool | dict = False,
        executor_cfg: str | dict | Any | None = None,
        with_adapter_cfg: dict | Any | None = None,
        pipeline_adapter_cfg: dict | Any | None = None,
        project_adapter_cfg: dict | Any | None = None,
        adapter: dict[str, Any] | None = None,
        reload: bool = False,
        log_level: str | None = None,
        max_retries: int | None = None,
        retry_delay: float | None = None,
        jitter_factor: float | None = None,
        retry_exceptions: tuple | None = None,
        # --- Schedule Parameters (passed to job queue.add_schedule) ---
        cron: str | dict[str, str | int] | None = None,
        interval: int | str | dict[str, str | int] | None = None,
        date: dt.datetime | None = None,
        overwrite: bool = False,
        schedule_id: str | None = None,
        **kwargs,
    ) -> str | UUID:
        """
        Schedule a pipeline for execution using the configured job queue.

        Args:
            run_func (Callable): The function to execute in the job queue.
            pipeline_cfg (PipelineConfig): The pipeline configuration object.
            inputs (dict | None): Inputs for the pipeline run (overrides config).
            final_vars (list | None): Final variables for the pipeline run (overrides config).
            config (dict | None): Hamilton driver config (overrides config).
            cache (bool | dict): Cache configuration (overrides config).
            executor_cfg (str | dict | ExecutorConfig | None): Executor configuration (overrides config).
            with_adapter_cfg (dict | WithAdapterConfig | None): Adapter configuration (overrides config).
            pipeline_adapter_cfg (dict | PipelineAdapterConfig | None): Pipeline adapter configuration (overrides config).
            project_adapter_cfg (dict | ProjectAdapterConfig | None): Project adapter configuration (overrides config).
            adapter (dict | None): Additional Hamilton adapters (overrides config).
            reload (bool): Whether to reload module (overrides config).
            log_level (str | None): Log level for the run (overrides config).
            max_retries (int): Maximum number of retries for the job.
            retry_delay (float): Delay between retries.
            jitter_factor (float): Jitter factor for retry delay.
            retry_exceptions (tuple): Exceptions that should trigger a retry.
            cron (str | dict | None): Cron expression or dict for cron trigger.
            interval (int | str | dict | None): Interval in seconds or dict for interval trigger.
            date (dt.datetime | None): Date for date trigger.
            overwrite (bool): If True and id_ is None, generates ID '{name}-1', potentially overwriting.
            schedule_id (str | None): Optional ID for the schedule. If None, generates a new ID.
            **kwargs: Additional keyword arguments passed to the job queue's add_schedule method,
                For RQ this includes:
                    - repeat: Repeat count (int or dict)
                    - result_ttl: Time to live for the job result (float or timedelta)
                    - ttl: Time to live for the job (float or timedelta)
                    - use_local_time_zone: Whether to use local time zone for scheduling (bool)
                For APScheduler, this includes:
                    - misfire_grace_time: Grace time for misfires (timedelta)
                    - coalesce: Whether to coalesce jobs (bool)
                    - max_running_jobs: Maximum instances of the job (int)
                    - max_jitter: Maximum jitter for scheduling (int)
                    - conflict_policy: Policy for conflicting jobs (str)
                    - paused: Whether to pause the job (bool)


        Returns:
            str | UUID: The ID of the scheduled pipeline.

        Raises:
            ValueError: If trigger_type is invalid or required args are missing.
            Exception: Can raise exceptions from the job queue backend.
        """

        project_name = self.project_cfg.name
        name = pipeline_cfg.name
        logger.debug(
            f"Attempting to schedule pipeline: {project_name}.{name} with id: {schedule_id}"
        )

        # --- Resolve Parameters using pipeline_cfg for defaults ---
        schedule_cfg = pipeline_cfg.schedule
        # run_cfg = pipeline_cfg.run

        pipeline_run_args = {
            "inputs": inputs,
            "final_vars": final_vars,
            "config": config,
            "cache": cache,
            "executor_cfg": executor_cfg,
            "with_adapter_cfg": with_adapter_cfg,
            "pipeline_adapter_cfg": pipeline_adapter_cfg,
            "project_adapter_cfg": project_adapter_cfg,
            "adapter": adapter,
            "reload": reload,
            "log_level": log_level,
            "max_retries": max_retries,
            "retry_delay": retry_delay,
            "jitter_factor": jitter_factor,
            "retry_exceptions": retry_exceptions,
        }
        pipeline_run_args = {
            k: v for k, v in pipeline_run_args.items() if v is not None
        }
        logger.debug(f"Resolved run_kwargs for '{name}': {pipeline_run_args}")

        cron = cron if cron is not None else schedule_cfg.cron
        interval = interval if interval is not None else schedule_cfg.interval
        date = date if date is not None else schedule_cfg.date
        logger.debug(
            f"Resolved schedule parameters for '{name}': cron={cron}, interval={interval}, date={date}"
        )

        # --- Generate ID if not provided ---
        # (Keep _generate_id function as is, it uses self._get_schedules())
        def _generate_id(
            pipeline_name: str, explicit_id: str | None, force_overwrite_base: bool
        ) -> str:
            if explicit_id:
                logger.debug(f"Using explicit schedule ID: {explicit_id}")
                return explicit_id

            base_id = f"{pipeline_name}-1"

            if force_overwrite_base:
                logger.debug(f"Overwrite specified, using base ID: {base_id}")
                return base_id

            try:
                existing_ids = self._get_schedule_ids()
                logger.debug(f"Existing schedule IDs: {existing_ids}")

                if not any(
                    id_val.startswith(f"{pipeline_name}-") for id_val in existing_ids
                ):
                    logger.debug(
                        f"No existing schedules found for '{pipeline_name}', using base ID: {base_id}"
                    )
                    return base_id

                # Find highest existing number for this pipeline name
                max_num = 0
                for id_val in existing_ids:
                    if id_val.startswith(f"{pipeline_name}-"):
                        try:
                            num_part = id_val.split("-")[-1]
                            num = int(num_part)
                            if num > max_num:
                                max_num = num
                        except (ValueError, IndexError):
                            logger.warning(
                                f"Could not parse number from existing schedule ID: {id_val}"
                            )
                            continue  # Skip malformed IDs

                new_id = f"{pipeline_name}-{max_num + 1}"
                logger.debug(f"Generated new schedule ID: {new_id}")
                return new_id

            except Exception as e:
                logger.error(
                    f"Error getting existing schedules to generate ID: {e}. Falling back to base ID: {base_id}"
                )
                # Fallback in case of error fetching schedules
                return base_id

        schedule_id = _generate_id(name, schedule_id, overwrite)

        # --- Add Schedule via Job queue ---
        try:
            with self.job_queue as job_queue:
                # Job queue is now responsible for creating the trigger object
                # Pass trigger type and kwargs directly
                added_id = job_queue.add_schedule(
                    func=run_func,
                    func_kwargs=pipeline_run_args,  # Pass resolved run parameters
                    cron=cron,
                    interval=interval,
                    date=date,
                    schedule_id=schedule_id,
                    **kwargs,  # Pass resolved schedule run parameters
                )
                logger.info(
                    f"✅ Successfully scheduled job for "
                    f"[blue]{project_name}.{name}[/blue] with ID [green]{added_id}[/green]"
                )
                return added_id
        except Exception as e:
            logger.error(
                f"Failed to add schedule '{schedule_id}' for pipeline '{name}': {e}"
            )
            raise

    # --- schedule_all method removed ---
    # PipelineManager will be responsible for iterating and calling schedule()

    def schedule_all(self, registry: PipelineRegistry, **kwargs):
        """
        Schedule all pipelines found by the registry.

        Args:
            **kwargs: Arguments passed directly to the `schedule` method for each pipeline.
                      Note: Pipeline-specific configurations will still take precedence for
                      defaults if not overridden by kwargs.
        """
        try:
            registry = self._get_registry_func()
            names = registry._get_names()  # Use registry to find pipelines
            if not names:
                logger.info("[yellow]No pipelines found to schedule.[/yellow]")
                return

            logger.info(f"Attempting to schedule {len(names)} pipelines...")
            scheduled_ids = []
            errors = []
            for name in names:
                try:
                    # Load config specifically for this pipeline to get defaults
                    # Note: schedule() will load it again, potential optimization later
                    cfg = self._load_config_func(name=name)
                    if (
                        not cfg
                        or not cfg.pipeline
                        or not cfg.pipeline.schedule
                        or not cfg.pipeline.schedule.enabled
                    ):
                        logger.info(
                            f"🟡 Skipping schedule for [cyan]{name}[/cyan]: Not configured or disabled in config."
                        )
                        continue

                    logger.info(f"Scheduling [cyan]{name}[/cyan]...")
                    # Pass kwargs, allowing overrides of config defaults
                    schedule_id = self.schedule(name=name, **kwargs)
                    scheduled_ids.append(schedule_id)
                except Exception as e:
                    logger.error(f"Failed to schedule pipeline '{name}': {e}")
                    errors.append(name)
                    logger.error(f"❌ Error scheduling [cyan]{name}[/cyan]: {e}")

            if errors:
                logger.error(
                    f"\n[bold red]Finished scheduling with errors for: {', '.join(errors)}[/bold red]"
                )
            else:
                logger.success(
                    f"\n[bold green]Successfully scheduled {len(scheduled_ids)} pipelines.[/bold green]"
                )

        except Exception as e:
            logger.error(
                f"[bold red]An unexpected error occurred during schedule_all: {e}[/bold red]"
            )



PIPELINE_PY_TEMPLATE = """# FlowerPower pipeline {name}.py
# Created on {date}

####################################################################################################
# Import necessary libraries
# NOTE: Remove or comment out imports that are not used in the pipeline

from hamilton.function_modifiers import parameterize, dataloader, datasaver
from hamilton.htypes import Parallelizable, Collect

from pathlib import Path

from flowerpower.cfg import Config

####################################################################################################
# Load pipeline parameters. Do not modify this section.

PARAMS = Config.load(
    Path(__file__).parents[1], pipeline_name="{name}"
).pipeline.h_params


####################################################################################################
# Helper functions.
# This functions have to start with an underscore (_).


####################################################################################################
# Pipeline functions

"""

HOOK_TEMPLATE__MQTT_BUILD_CONFIG = '''
def {function_name}(payload: bytes, topic: str) -> dict:
    """
    MQTT hook function to build the configuration for the pipeline.
    This function is called in the on_message callback of the MQTT client.
    The result of this function will be passed to the hamilton builder as the config for the pipeline.
    Args:
        payload (bytes): The payload of the MQTT message.
        topic (str): The topic of the MQTT message.
    Returns:
        dict: The configuration for the pipeline.
    """

    pass
'''



from ..base import BaseDatasetWriter


class PydalaDatasetWriter(BaseDatasetWriter):
    """Writer for Pydala dataset.

    This class is responsible for writing dataframes to Pydala dataset.

    Examples:
        ```python
        writer = PydalaDatasetWriter(path="pydala_data/")
        writer.write(df)
        ```
    """

    format: str = "parquet"
    is_pydala_dataset: bool = True

    def model_post_init(self, __context):
        super().model_post_init(__context)
